# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains dialogues used by the coin_price skill."""

from typing import Any

from aea.protocols.base import Address, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.skills.base import Model

from packages.fetchai.protocols.http.dialogues import HttpDialogue as BaseHttpDialogue
from packages.fetchai.protocols.http.dialogues import HttpDialogues as BaseHttpDialogues
from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.protocols.prometheus.dialogues import (
    PrometheusDialogue as BasePrometheusDialogue,
)
from packages.fetchai.protocols.prometheus.dialogues import (
    PrometheusDialogues as BasePrometheusDialogues,
)


HttpDialogue = BaseHttpDialogue
PrometheusDialogue = BasePrometheusDialogue


class HttpDialogues(Model, BaseHttpDialogues):
    """This class keeps track of all http dialogues."""

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize dialogues.

        :param agent_address: the address of the agent for whom dialogues are maintained
        :return: None
        """
        Model.__init__(self, **kwargs)

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent in this dialogue
            """
            if (
                message.performative == HttpMessage.Performative.REQUEST
                and message.sender != receiver_address
            ) or (
                message.performative == HttpMessage.Performative.RESPONSE
                and message.sender == receiver_address
            ):
                return BaseHttpDialogue.Role.SERVER

            return BaseHttpDialogue.Role.CLIENT

        BaseHttpDialogues.__init__(
            self,
            self_address=str(self.skill_id),
            role_from_first_message=role_from_first_message,
        )


class PrometheusDialogues(Model, BasePrometheusDialogues):
    """The dialogues class keeps track of all prometheus dialogues."""

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        self.enabled = kwargs.pop("enabled", False)
        self.metrics = kwargs.pop("metrics", [])

        Model.__init__(self, **kwargs)

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return PrometheusDialogue.Role.AGENT

        BasePrometheusDialogues.__init__(
            self,
            self_address=str(self.skill_id),
            role_from_first_message=role_from_first_message,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the implementation of AEA agents manager."""
import asyncio
import json
import os
import threading
from asyncio.tasks import FIRST_COMPLETED
from collections import defaultdict
from shutil import rmtree
from threading import Thread
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

from aea.aea import AEA
from aea.configurations.constants import AEA_MANAGER_DATA_DIRNAME, DEFAULT_REGISTRY_NAME
from aea.configurations.data_types import PublicId
from aea.helpers.io import open_file
from aea.manager.project import AgentAlias, Project


class ProjectNotFoundError(ValueError):
    """Project not found exception."""


class ProjectCheckError(ValueError):
    """Project check error exception."""

    def __init__(self, msg: str, source_exception: Exception):
        """Init exception."""
        super().__init__(msg)
        self.source_exception = source_exception


class AgentRunAsyncTask:
    """Async task wrapper for agent."""

    def __init__(self, agent: AEA, loop: asyncio.AbstractEventLoop) -> None:
        """Init task with agent and loop."""
        self.run_loop: asyncio.AbstractEventLoop = loop
        self.caller_loop: asyncio.AbstractEventLoop = loop
        self._done_future: Optional[asyncio.Future] = None
        self.task: Optional[asyncio.Task] = None
        self.agent = agent

    def create_run_loop(self) -> None:
        """Create run loop."""

    def start(self) -> None:
        """Start task."""
        self.create_run_loop()
        self.task = self.run_loop.create_task(self._run_wrapper())
        self._done_future = asyncio.Future(loop=self.caller_loop)

    def wait(self) -> asyncio.Future:
        """Return future to wait task completed."""
        if not self._done_future:  # pragma: nocover
            raise ValueError("Task not started!")
        return self._done_future

    def stop(self) -> None:
        """Stop task."""
        if not self.run_loop or not self.task:  # pragma: nocover
            raise ValueError("Task was not started!")
        self.run_loop.call_soon_threadsafe(self.task.cancel)

    async def _run_wrapper(self) -> None:
        """Run task internals."""
        if not self._done_future:  # pragma: nocover
            raise ValueError("Task was not started! please use start method")
        exc = None
        try:
            await self.run()
        except asyncio.CancelledError:  # pragma: nocover
            pass
        except Exception as e:  # pylint: disable=broad-except
            exc = e
        finally:
            self.caller_loop.call_soon_threadsafe(self._set_result, exc)

    def _set_result(self, exc: Optional[BaseException]) -> None:
        """Set result of task execution."""
        if not self._done_future or self._done_future.done():  # pragma: nocover
            return
        if exc:
            self._done_future.set_exception(exc)
        else:
            self._done_future.set_result(None)

    async def run(self) -> None:
        """Run task body."""
        self.agent.runtime.set_loop(self.run_loop)
        await self.agent.runtime.run()

    @property
    def is_running(self) -> bool:
        """Return is task running."""
        return not self.wait().done()


class AgentRunThreadTask(AgentRunAsyncTask):
    """Threaded wrapper to run agent."""

    def __init__(self, agent: AEA, loop: asyncio.AbstractEventLoop) -> None:
        """Init task with agent and loop."""
        AgentRunAsyncTask.__init__(self, agent, loop)
        self._thread: Optional[Thread] = None

    def create_run_loop(self) -> None:
        """Create run loop."""
        self.run_loop = asyncio.new_event_loop()

    def start(self) -> None:
        """Run task in a dedicated thread."""
        super().start()
        self._thread = threading.Thread(
            target=self.run_loop.run_until_complete, args=[self.task], daemon=True
        )
        self._thread.start()

    def stop(self,) -> None:
        """Stop the task."""
        super().stop()
        if self._thread is not None:
            self._thread.join()


class MultiAgentManager:
    """Multi agents manager."""

    MODES = ["async", "threaded"]
    DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS = 60
    SAVE_FILENAME = "save.json"

    def __init__(
        self,
        working_dir: str,
        mode: str = "async",
        registry_path: str = DEFAULT_REGISTRY_NAME,
        auto_add_remove_project: bool = False,
    ) -> None:
        """
        Initialize manager.

        :param working_dir: directory to store base agents.
        :param mode: str. async or threaded
        :param registry_path: str. path to the local packages registry
        :param auto_add_remove_project: bool. add/remove project on the first agent add/last agent remove

        :return: None
        """
        self.working_dir = working_dir
        self._auto_add_remove_project = auto_add_remove_project
        self._save_path = os.path.join(self.working_dir, self.SAVE_FILENAME)

        self.registry_path = registry_path
        self._was_working_dir_created = False
        self._is_running = False
        self._projects: Dict[PublicId, Project] = {}
        self._versionless_projects_set: Set[PublicId] = set()
        self._data_dir = os.path.abspath(
            os.path.join(self.working_dir, AEA_MANAGER_DATA_DIRNAME)
        )
        self._agents: Dict[str, AgentAlias] = {}
        self._agents_tasks: Dict[str, AgentRunAsyncTask] = {}

        self._thread: Optional[Thread] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._event: Optional[asyncio.Event] = None

        self._error_callbacks: List[Callable[[str, BaseException], None]] = []
        self._last_start_status: Optional[
            Tuple[
                bool,
                Dict[PublicId, List[Dict]],
                List[Tuple[PublicId, List[Dict], Exception]],
            ]
        ] = None

        if mode not in self.MODES:
            raise ValueError(
                f'Invalid mode {mode}. Valid modes are {", ".join(self.MODES)}'
            )
        self._started_event = threading.Event()
        self._mode = mode

    @property
    def data_dir(self) -> str:
        """Get the certs directory."""
        return self._data_dir

    def get_data_dir_of_agent(self, agent_name: str) -> str:
        """Get the data directory of a specific agent."""
        return os.path.join(self.data_dir, agent_name)

    @property
    def is_running(self) -> bool:
        """Is manager running."""
        return self._is_running

    @property
    def dict_state(self) -> Dict[str, Any]:
        """Create MultiAgentManager dist state."""
        return {
            "projects": [str(public_id) for public_id in self._projects.keys()],
            "agents": [alias.dict for alias in self._agents.values()],
        }

    def _run_thread(self) -> None:
        """Run internal thread with own event loop."""
        self._loop = asyncio.new_event_loop()
        self._event = asyncio.Event(loop=self._loop)
        self._loop.run_until_complete(self._manager_loop())

    async def _manager_loop(self) -> None:
        """Await and control running manager."""
        if not self._event:  # pragma: nocover
            raise ValueError("Do not use this method directly, use start_manager.")

        self._started_event.set()

        while self._is_running:
            agents_run_tasks_futures = {
                task.wait(): agent_name
                for agent_name, task in self._agents_tasks.items()
            }
            wait_tasks = list(agents_run_tasks_futures.keys()) + [self._event.wait()]  # type: ignore
            done, _ = await asyncio.wait(wait_tasks, return_when=FIRST_COMPLETED)

            if self._event.is_set():
                self._event.clear()

            for task in done:
                if task not in agents_run_tasks_futures:
                    # task not in agents_run_tasks_futures, so it's event_wait, skip it
                    await task
                    continue

                agent_name = agents_run_tasks_futures[task]
                self._agents_tasks.pop(agent_name)
                if task.exception():
                    for callback in self._error_callbacks:
                        callback(agent_name, task.exception())
                else:
                    await task

    def add_error_callback(
        self, error_callback: Callable[[str, BaseException], None]
    ) -> None:
        """Add error callback to call on error raised."""
        self._error_callbacks.append(error_callback)

    def start_manager(
        self, local: bool = False, remote: bool = False
    ) -> "MultiAgentManager":
        """
        Start manager.

        If local = False and remote = False, then the packages
        are fetched in mixed mode (i.e. first try from local
        registry, and then from remote registry in case of failure).

        :param local: whether or not to fetch from local registry.
        :param remote: whether or not to fetch from remote registry.

        :return: the MultiAgentManager instance.
        """
        if self._is_running:
            return self

        self._ensure_working_dir()
        self._last_start_status = self._load_state(local=local, remote=remote)

        self._started_event.clear()
        self._is_running = True
        self._thread = Thread(target=self._run_thread, daemon=True)
        self._thread.start()
        self._started_event.wait(self.DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS)
        return self

    @property
    def last_start_status(
        self,
    ) -> Tuple[
        bool, Dict[PublicId, List[Dict]], List[Tuple[PublicId, List[Dict], Exception]],
    ]:
        """Get status of the last agents start loading state."""
        if self._last_start_status is None:
            raise ValueError("Manager was not started")
        return self._last_start_status

    def stop_manager(
        self, cleanup: bool = True, save: bool = False
    ) -> "MultiAgentManager":
        """
        Stop manager.

        Stops all running agents and stop agent.

        :param cleanup: bool is cleanup on stop.
        :param save: bool is save state to file on stop.

        :return: None
        """
        if not self._is_running:
            return self

        if not self._loop or not self._event or not self._thread:  # pragma: nocover
            raise ValueError("Manager was not started!")

        if not self._thread.is_alive():  # pragma: nocover
            return self

        self.stop_all_agents()

        if save:
            self._save_state()

        for agent_name in self.list_agents():
            self.remove_agent(agent_name, skip_project_auto_remove=True)

        if cleanup:
            for project in list(self._projects.keys()):
                self.remove_project(project, keep_files=save)
            self._cleanup(only_data=save)

        self._is_running = False

        self._loop.call_soon_threadsafe(self._event.set)

        if self._thread.ident != threading.get_ident():
            self._thread.join(self.DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS)

        self._thread = None
        return self

    def _cleanup(self, only_data: bool = False) -> None:
        """Remove workdir if was created."""
        if only_data:
            rmtree(self.data_dir)
        else:
            if self._was_working_dir_created and os.path.exists(self.working_dir):
                rmtree(self.working_dir)

    def add_project(
        self,
        public_id: PublicId,
        local: bool = False,
        remote: bool = False,
        restore: bool = False,
    ) -> "MultiAgentManager":
        """
        Fetch agent project and all dependencies to working_dir.

        If local = False and remote = False, then the packages
        are fetched in mixed mode (i.e. first try from local
        registry, and then from remote registry in case of failure).

        :param public_id: the public if of the agent project.

        :param local: whether or not to fetch from local registry.
        :param remote: whether or not to fetch from remote registry.
        :param restore: bool flag for restoring already fetched agent.
        """
        if public_id.to_any() in self._versionless_projects_set:
            raise ValueError(
                f"The project ({public_id.author}/{public_id.name}) was already added!"
            )

        self._versionless_projects_set.add(public_id.to_any())

        project = Project.load(
            self.working_dir,
            public_id,
            local,
            remote,
            registry_path=self.registry_path,
            is_restore=restore,
        )

        if not restore:
            project.install_pypi_dependencies()
            project.build()

        try:
            project.check()
        except Exception as e:
            project.remove()
            raise ProjectCheckError(
                f"Failed to load project: {public_id} Error: {str(e)}", e
            )

        self._projects[public_id] = project
        return self

    def remove_project(
        self, public_id: PublicId, keep_files: bool = False
    ) -> "MultiAgentManager":
        """Remove agent project."""
        if public_id not in self._projects:
            raise ValueError(f"Project {public_id} is not present!")

        if self._projects[public_id].agents:
            raise ValueError(
                f"Can not remove projects with aliases exists: {self._projects[public_id].agents}"
            )

        project = self._projects.pop(public_id)
        self._versionless_projects_set.remove(public_id.to_any())
        if not keep_files:
            project.remove()

        return self

    def list_projects(self) -> List[PublicId]:
        """
        List all agents projects added.

        :return: list of public ids of projects
        """
        return list(self._projects.keys())

    def add_agent(
        self,
        public_id: PublicId,
        agent_name: Optional[str] = None,
        agent_overrides: Optional[dict] = None,
        component_overrides: Optional[List[dict]] = None,
        local: bool = False,
        remote: bool = False,
        restore: bool = False,
    ) -> "MultiAgentManager":
        """
        Create new agent configuration based on project with config overrides applied.

        Alias is stored in memory only!

        :param public_id: base agent project public id
        :param agent_name: unique name for the agent
        :param agent_overrides: overrides for agent config.
        :param component_overrides: overrides for component section.
        :param config: agent config (used for agent re-creation).

        :param local: whether or not to fetch from local registry.
        :param remote: whether or not to fetch from remote registry.
        :param restore: bool flag for restoring already fetched agent.

        :return: manager
        """
        agent_name = agent_name or public_id.name

        if agent_name in self._agents:
            raise ValueError(f"Agent with name {agent_name} already exists!")

        project = self._projects.get(public_id, None)

        if project is None and self._auto_add_remove_project:
            self.add_project(public_id, local, remote, restore)
            project = self._projects.get(public_id, None)

        if project is None:
            raise ProjectNotFoundError(f"{public_id} project is not added!")

        agent_alias = AgentAlias(
            project=project,
            agent_name=agent_name,
            data_dir=self.get_data_dir_of_agent(agent_name),
        )
        agent_alias.set_overrides(agent_overrides, component_overrides)
        project.agents.add(agent_name)
        self._agents[agent_name] = agent_alias
        return self

    def add_agent_with_config(
        self, public_id: PublicId, config: List[dict], agent_name: Optional[str] = None,
    ) -> "MultiAgentManager":
        """
        Create new agent configuration based on project with config provided.

        Alias is stored in memory only!

        :param public_id: base agent project public id
        :param agent_name: unique name for the agent
        :param config: agent config (used for agent re-creation).

        :return: manager
        """
        agent_name = agent_name or public_id.name

        if agent_name in self._agents:  # pragma: nocover
            raise ValueError(f"Agent with name {agent_name} already exists!")

        if public_id not in self._projects:  # pragma: nocover
            raise ValueError(f"{public_id} project is not added!")

        project = self._projects[public_id]

        agent_alias = AgentAlias(
            project=project,
            agent_name=agent_name,
            data_dir=self.get_data_dir_of_agent(agent_name),
        )
        agent_alias.set_agent_config_from_data(config)
        project.agents.add(agent_name)
        self._agents[agent_name] = agent_alias
        return self

    def get_agent_overridables(self, agent_name: str) -> Tuple[Dict, List[Dict]]:
        """
        Get agent config  overridables.

        :param agent_name: str

        :return: Tuple of agent overridables dict and  and list of component overridables dict.
        """
        if agent_name not in self._agents:  # pragma: nocover
            raise ValueError(f"Agent with name {agent_name} does not exist!")

        return self._agents[agent_name].get_overridables()

    def set_agent_overrides(
        self,
        agent_name: str,
        agent_overides: Optional[Dict],
        components_overrides: Optional[List[Dict]],
    ) -> None:
        """
        Set agent overrides.

        :param agent_name: str
        :param agent_overides: optional dict of agent config overrides
        :param components_overrides: optional list of dict of components overrides

        :return: None
        """
        if agent_name not in self._agents:  # pragma: nocover
            raise ValueError(f"Agent with name {agent_name} does not exist!")

        if self._is_agent_running(agent_name):  # pragma: nocover
            raise ValueError("Agent is running. stop it first!")

        self._agents[agent_name].set_overrides(agent_overides, components_overrides)

    def list_agents_info(self) -> List[Dict[str, Any]]:
        """
        List agents detailed info.

        :return: list of dicts that represents agent info: public_id, name, is_running.
        """
        return [
            {
                "agent_name": agent_name,
                "public_id": str(alias.project.public_id),
                "addresses": alias.get_addresses(),
                "is_running": self._is_agent_running(agent_name),
            }
            for agent_name, alias in self._agents.items()
        ]

    def list_agents(self, running_only: bool = False) -> List[str]:
        """
        List all agents.

        :param running_only: returns only running if set to True

        :return: list of agents names
        """
        if running_only:
            return [i for i in self._agents.keys() if self._is_agent_running(i)]
        return list(self._agents.keys())

    def remove_agent(
        self, agent_name: str, skip_project_auto_remove: bool = False
    ) -> "MultiAgentManager":
        """
        Remove agent alias definition from registry.

        :param agent_name: agent name to remove
        :param skip_project_auto_remove: disable auto project remove on last agent removed.

        :return: None
        """
        if agent_name not in self._agents:
            raise ValueError(f"Agent with name {agent_name} does not exist!")

        if self._is_agent_running(agent_name):
            raise ValueError("Agent is running. stop it first!")

        agent_alias = self._agents.pop(agent_name)
        agent_alias.remove_from_project()
        project: Project = agent_alias.project

        if (
            not project.agents
            and self._auto_add_remove_project
            and not skip_project_auto_remove
        ):
            self.remove_project(project.public_id, keep_files=False)

        return self

    def start_agent(self, agent_name: str) -> "MultiAgentManager":
        """
        Start selected agent.

        :param agent_name: agent name to start

        :return: None
        """
        if not self._loop or not self._event:  # pragma: nocover
            raise ValueError("agent is not started!")

        agent_alias = self._agents.get(agent_name)

        if not agent_alias:
            raise ValueError(f"{agent_name} is not registered!")

        if self._is_agent_running(agent_name):
            raise ValueError(f"{agent_name} is already started!")

        aea = agent_alias.get_aea_instance()

        if self._mode == "async":
            task = AgentRunAsyncTask(aea, self._loop)
        elif self._mode == "threaded":
            task = AgentRunThreadTask(aea, self._loop)

        task.start()
        self._agents_tasks[agent_name] = task
        self._loop.call_soon_threadsafe(self._event.set)
        return self

    def _is_agent_running(self, agent_name: str) -> bool:
        """Return is agent running state."""
        if agent_name not in self._agents_tasks:
            return False

        task = self._agents_tasks[agent_name]
        return task.is_running

    def start_all_agents(self) -> "MultiAgentManager":
        """
        Start all not started agents.

        :return: None
        """
        self.start_agents(
            [
                agent_name
                for agent_name in self.list_agents()
                if not self._is_agent_running(agent_name)
            ]
        )

        return self

    def stop_agent(self, agent_name: str) -> "MultiAgentManager":
        """
        Stop running agent.

        :param agent_name: agent name to stop

        :return: None
        """
        if not self._is_agent_running(agent_name) or not self._thread or not self._loop:
            raise ValueError(f"{agent_name} is not running!")

        agent_task = self._agents_tasks[agent_name]

        if self._thread.ident == threading.get_ident():  # pragma: nocover
            # In same thread do not perform blocking operations!
            agent_task.stop()
            return self

        wait_future = agent_task.wait()

        event = threading.Event()

        def event_set(*args: Any) -> None:  # pylint: disable=unused-argument
            event.set()

        def _add_cb() -> None:
            if wait_future.done():
                event_set()  # pragma: nocover
            else:
                wait_future.add_done_callback(event_set)  # pramga: nocover

        self._loop.call_soon_threadsafe(_add_cb)
        agent_task.stop()
        event.wait(self.DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS)

        return self

    def stop_all_agents(self) -> "MultiAgentManager":
        """
        Stop all agents running.

        :return: None
        """
        agents_list = self.list_agents(running_only=True)
        self.stop_agents(agents_list)

        return self

    def stop_agents(self, agent_names: List[str]) -> "MultiAgentManager":
        """
        Stop specified agents.

        :return: None
        """
        for agent_name in agent_names:
            if not self._is_agent_running(agent_name):
                raise ValueError(f"{agent_name} is not running!")

        for agent_name in agent_names:
            self.stop_agent(agent_name)

        return self

    def start_agents(self, agent_names: List[str]) -> "MultiAgentManager":
        """
        Stop specified agents.

        :return: None
        """
        for agent_name in agent_names:
            self.start_agent(agent_name)

        return self

    def get_agent_alias(self, agent_name: str) -> AgentAlias:
        """
        Return details about agent alias definition.

        :return: AgentAlias
        """
        if agent_name not in self._agents:  # pragma: nocover
            raise ValueError(f"Agent with name {agent_name} does not exist!")
        return self._agents[agent_name]

    def _ensure_working_dir(self) -> None:
        """Create working dir if needed."""
        if not os.path.exists(self.working_dir):
            os.makedirs(self.working_dir)
            self._was_working_dir_created = True

        if not os.path.isdir(self.working_dir):  # pragma: nocover
            raise ValueError(f"{self.working_dir} is not a directory!")
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)

    def _load_state(
        self, local: bool, remote: bool
    ) -> Tuple[
        bool, Dict[PublicId, List[Dict]], List[Tuple[PublicId, List[Dict], Exception]],
    ]:
        """
        Load saved state from file.

        Fetch agent project and all dependencies to working_dir.

        If local = False and remote = False, then the packages
        are fetched in mixed mode (i.e. first try from local
        registry, and then from remote registry in case of failure).

        :param local: whether or not to fetch from local registry.
        :param remote: whether or not to fetch from remote registry.

        :return: None
        :raises: ValueError if failed to load state.
        """
        if not os.path.exists(self._save_path):
            return False, {}, []

        save_json = {}
        with open_file(self._save_path) as f:
            save_json = json.load(f)

        if not save_json:
            return False, {}, []  # pragma: nocover

        projects_agents: Dict[PublicId, List] = defaultdict(list)

        for agent_settings in save_json["agents"]:
            projects_agents[PublicId.from_str(agent_settings["public_id"])].append(
                agent_settings
            )

        failed_to_load: List[Tuple[PublicId, List[Dict], Exception]] = []
        loaded_ok: Dict[PublicId, List[Dict]] = {}
        for project_public_id, agents_settings in projects_agents.items():
            try:
                self.add_project(
                    project_public_id, local=local, remote=remote, restore=True,
                )
            except ProjectCheckError as e:
                failed_to_load.append((project_public_id, agents_settings, e))
                break

            for agent_settings in agents_settings:

                self.add_agent_with_config(
                    public_id=PublicId.from_str(agent_settings["public_id"]),
                    agent_name=agent_settings["agent_name"],
                    config=agent_settings["config"],
                )

            loaded_ok[project_public_id] = agents_settings

        return True, loaded_ok, failed_to_load

    def _save_state(self) -> None:
        """
        Save MultiAgentManager state.

        :return: None.
        """
        with open_file(self._save_path, "w") as f:
            json.dump(self.dict_state, f, indent=4, sort_keys=True)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea generate' subcommand."""

import os
from typing import Set

import click
import yaml

from aea.cli.fingerprint import fingerprint_item
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project, clean_after, pass_ctx
from aea.cli.utils.loggers import logger
from aea.configurations.base import (
    ProtocolSpecification,
    ProtocolSpecificationParseError,
    PublicId,
)
from aea.configurations.constants import (
    DEFAULT_AEA_CONFIG_FILE,
    PROTOCOL,
    PROTOCOL_LANGUAGE_PYTHON,
    SUPPORTED_PROTOCOL_LANGUAGES,
)
from aea.helpers.io import open_file
from aea.protocols.generator.base import ProtocolGenerator


@click.group()
@click.pass_context
@check_aea_project
def generate(
    click_context: click.core.Context,  # pylint: disable=unused-argument
) -> None:
    """Generate a package for the agent."""


@generate.command()
@click.argument("protocol_specification_path", type=str, required=True)
@click.option(
    "--l",
    "language",
    type=click.Choice(SUPPORTED_PROTOCOL_LANGUAGES),
    required=False,
    default=PROTOCOL_LANGUAGE_PYTHON,
    help="Specify the language in which to generate the protocol package.",
)
@pass_ctx
def protocol(ctx: Context, protocol_specification_path: str, language: str) -> None:
    """Generate a protocol based on a specification and add it to the configuration file and agent."""
    ctx.set_config("language", language)
    _generate_protocol(ctx, protocol_specification_path)


@clean_after
def _generate_protocol(ctx: Context, protocol_specification_path: str) -> None:
    """Generate a protocol based on a specification and add it to the configuration file and agent."""
    protocol_plural = PROTOCOL + "s"

    # Create protocol generator (load, validate,
    # extract fields from protocol specification yaml file)
    try:
        output_path = os.path.join(ctx.cwd, protocol_plural)
        protocol_generator = ProtocolGenerator(protocol_specification_path, output_path)
    except FileNotFoundError as e:
        raise click.ClickException(  # pragma: no cover
            "Protocol is NOT generated. The following error happened while generating the protocol:\n"
            + str(e)
        )
    except yaml.YAMLError as e:
        raise click.ClickException(  # pragma: no cover
            "Protocol is NOT generated. The following error happened while generating the protocol:\n"
            + "Yaml error in the protocol specification file:"
            + str(e)
        )
    except ProtocolSpecificationParseError as e:
        raise click.ClickException(  # pragma: no cover
            "Protocol is NOT generated. The following error happened while generating the protocol:\n"
            + "Error while parsing the protocol specification: "
            + str(e)
        )
    except Exception as e:  # pragma: no cover
        raise click.ClickException(  # pragma: no cover
            "Protocol is NOT generated. The following error happened while generating the protocol:\n"
            + str(e)
        )

    # helpers
    language = ctx.config.get("language")
    existing_protocol_ids_list = getattr(ctx.agent_config, "{}s".format(PROTOCOL))
    existing_protocol_name_list = [
        public_id.name for public_id in existing_protocol_ids_list
    ]
    protocol_spec = protocol_generator.protocol_specification
    protocol_directory_path = os.path.join(ctx.cwd, protocol_plural, protocol_spec.name)

    # Check if a protocol with the same name exists in the agent config
    if protocol_spec.name in existing_protocol_name_list:
        raise click.ClickException(
            "Protocol is NOT generated. The following error happened while generating the protocol:\n"
            + f"A {PROTOCOL} with name '{protocol_spec.name}' already exists. Aborting..."
        )

    # Check if a directory with the same name as the protocol's exists
    # in the protocols directory of the agent's directory
    if os.path.exists(protocol_directory_path):
        raise click.ClickException(
            "Protocol is NOT generated. The following error happened while generating the protocol:\n"
            + f"A directory with name '{protocol_spec.name}' already exists. Aborting..."
        )

    ctx.clean_paths.append(protocol_directory_path)
    agent_name = ctx.agent_config.agent_name
    click.echo(
        "Generating {} '{}' and adding it to the agent '{}'...".format(
            PROTOCOL, protocol_spec.name, agent_name
        )
    )

    if language == PROTOCOL_LANGUAGE_PYTHON:
        _generate_full_mode(
            ctx, protocol_generator, protocol_spec, existing_protocol_ids_list, language
        )
    else:
        _generate_protobuf_mode(ctx, protocol_generator, language)


@clean_after
def _generate_full_mode(
    ctx: Context,
    protocol_generator: ProtocolGenerator,
    protocol_spec: ProtocolSpecification,
    existing_id_list: Set[PublicId],
    language: str,
) -> None:
    """Generate a protocol in 'full' mode, and add it to the configuration file and agent."""
    try:
        warning_message = protocol_generator.generate(
            protobuf_only=False, language=language
        )
        if warning_message is not None:
            click.echo(warning_message)

        # Add the item to the configurations
        logger.debug(
            "Registering the {} into {}".format(PROTOCOL, DEFAULT_AEA_CONFIG_FILE)
        )
        existing_id_list.add(
            PublicId(protocol_spec.author, protocol_spec.name, protocol_spec.version)
        )
        ctx.agent_loader.dump(
            ctx.agent_config,
            open_file(os.path.join(ctx.cwd, DEFAULT_AEA_CONFIG_FILE), "w"),
        )
    except FileExistsError:
        raise click.ClickException(  # pragma: no cover
            "A {} with this name already exists. Please choose a different name and try again.".format(
                PROTOCOL
            )
        )
    except Exception as e:
        raise click.ClickException(
            "Protocol is NOT generated. The following error happened while generating the protocol:\n"
            + str(e)
        )
    fingerprint_item(ctx, PROTOCOL, protocol_spec.public_id)


@clean_after
def _generate_protobuf_mode(
    ctx: Context,  # pylint: disable=unused-argument
    protocol_generator: ProtocolGenerator,
    language: str,
) -> None:
    """Generate a protocol in 'protobuf' mode, and add it to the configuration file and agent."""
    try:
        warning_message = protocol_generator.generate(
            protobuf_only=True, language=language
        )
        if warning_message is not None:
            click.echo(warning_message)
    except FileExistsError:  # pragma: no cover
        raise click.ClickException(  # pragma: no cover
            f"A {PROTOCOL} with this name already exists. Please choose a different name and try again."
        )
    except Exception as e:  # pragma: no cover
        raise click.ClickException(  # pragma: no cover
            "Protocol is NOT generated. The following error happened while generating the protocol:\n"
            + str(e)
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the simple_data_request skill."""
from pathlib import Path

from aea.test_tools.test_skill import BaseSkillTestCase

from tests.conftest import ROOT_DIR


class ConfirmationAW3TestCase(BaseSkillTestCase):
    """Sets the confirmation_aw3 class up for testing (overrides the necessary config values so tests can be done on confirmation_aw3 skill)."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "confirmation_aw3")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.aw1_aea = "some_aw1_aea"
        cls.leaderboard_url = "some_leaderboard_url"
        cls.leaderboard_token = "some_leaderboard_token"
        config_overrides = {
            "models": {
                "strategy": {
                    "args": {
                        "aw1_aea": cls.aw1_aea,
                        "leaderboard_url": cls.leaderboard_url,
                        "leaderboard_token": cls.leaderboard_token,
                    }
                }
            }
        }

        super().setup(config_overrides=config_overrides)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Serialization module for yoti protocol."""

from typing import Any, Dict, cast

from aea.mail.base_pb2 import DialogueMessage
from aea.mail.base_pb2 import Message as ProtobufMessage
from aea.protocols.base import Message, Serializer

from packages.fetchai.protocols.yoti import yoti_pb2
from packages.fetchai.protocols.yoti.message import YotiMessage


class YotiSerializer(Serializer):
    """Serialization for the 'yoti' protocol."""

    @staticmethod
    def encode(msg: Message) -> bytes:
        """
        Encode a 'Yoti' message into bytes.

        :param msg: the message object.
        :return: the bytes.
        """
        msg = cast(YotiMessage, msg)
        message_pb = ProtobufMessage()
        dialogue_message_pb = DialogueMessage()
        yoti_msg = yoti_pb2.YotiMessage()

        dialogue_message_pb.message_id = msg.message_id
        dialogue_reference = msg.dialogue_reference
        dialogue_message_pb.dialogue_starter_reference = dialogue_reference[0]
        dialogue_message_pb.dialogue_responder_reference = dialogue_reference[1]
        dialogue_message_pb.target = msg.target

        performative_id = msg.performative
        if performative_id == YotiMessage.Performative.GET_PROFILE:
            performative = yoti_pb2.YotiMessage.Get_Profile_Performative()  # type: ignore
            token = msg.token
            performative.token = token
            dotted_path = msg.dotted_path
            performative.dotted_path = dotted_path
            args = msg.args
            performative.args.extend(args)
            yoti_msg.get_profile.CopyFrom(performative)
        elif performative_id == YotiMessage.Performative.PROFILE:
            performative = yoti_pb2.YotiMessage.Profile_Performative()  # type: ignore
            info = msg.info
            performative.info.update(info)
            yoti_msg.profile.CopyFrom(performative)
        elif performative_id == YotiMessage.Performative.ERROR:
            performative = yoti_pb2.YotiMessage.Error_Performative()  # type: ignore
            error_code = msg.error_code
            performative.error_code = error_code
            error_msg = msg.error_msg
            performative.error_msg = error_msg
            yoti_msg.error.CopyFrom(performative)
        else:
            raise ValueError("Performative not valid: {}".format(performative_id))

        dialogue_message_pb.content = yoti_msg.SerializeToString()

        message_pb.dialogue_message.CopyFrom(dialogue_message_pb)
        message_bytes = message_pb.SerializeToString()
        return message_bytes

    @staticmethod
    def decode(obj: bytes) -> Message:
        """
        Decode bytes into a 'Yoti' message.

        :param obj: the bytes object.
        :return: the 'Yoti' message.
        """
        message_pb = ProtobufMessage()
        yoti_pb = yoti_pb2.YotiMessage()
        message_pb.ParseFromString(obj)
        message_id = message_pb.dialogue_message.message_id
        dialogue_reference = (
            message_pb.dialogue_message.dialogue_starter_reference,
            message_pb.dialogue_message.dialogue_responder_reference,
        )
        target = message_pb.dialogue_message.target

        yoti_pb.ParseFromString(message_pb.dialogue_message.content)
        performative = yoti_pb.WhichOneof("performative")
        performative_id = YotiMessage.Performative(str(performative))
        performative_content = dict()  # type: Dict[str, Any]
        if performative_id == YotiMessage.Performative.GET_PROFILE:
            token = yoti_pb.get_profile.token
            performative_content["token"] = token
            dotted_path = yoti_pb.get_profile.dotted_path
            performative_content["dotted_path"] = dotted_path
            args = yoti_pb.get_profile.args
            args_tuple = tuple(args)
            performative_content["args"] = args_tuple
        elif performative_id == YotiMessage.Performative.PROFILE:
            info = yoti_pb.profile.info
            info_dict = dict(info)
            performative_content["info"] = info_dict
        elif performative_id == YotiMessage.Performative.ERROR:
            error_code = yoti_pb.error.error_code
            performative_content["error_code"] = error_code
            error_msg = yoti_pb.error.error_msg
            performative_content["error_msg"] = error_msg
        else:
            raise ValueError("Performative not valid: {}.".format(performative_id))

        return YotiMessage(
            message_id=message_id,
            dialogue_reference=dialogue_reference,
            target=target,
            performative=performative,
            **performative_content
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for aea helpers storage code."""
import os
import time

import pytest

from aea.helpers.storage.generic_storage import Storage


class TestAsyncCollection:
    """Test async storage collection."""

    @pytest.mark.asyncio
    async def test_collection(self):
        """Test collecton methods."""
        s = Storage("sqlite://:memory:")
        s.start()

        await s.wait_connected()

        col = await s.get_collection("test_col")
        col2 = await s.get_collection("another_collection")
        obj_id = "1"
        obj_body = {"a": 12}
        await col.put(obj_id, {"x": 13})
        await col.put(obj_id, obj_body)
        assert await col.find("a", 12) == [(obj_id, obj_body)]
        assert await col.get(obj_id) == obj_body
        assert await col2.get(obj_id) is None
        assert await col.get("not exists") is None

        assert await col.list() == [(obj_id, obj_body)]

        await col.remove(obj_id)
        assert await col.get(obj_id) is None

        s.stop()
        await s.wait_completed()


class TestSyncCollection:
    """Test sync storage collection."""

    def test_collection(self):
        """Test collecton methods."""
        s = Storage("sqlite://:memory:", threaded=True)
        s.start()

        while not s.is_connected:
            time.sleep(0.01)

        obj_id = "1"
        obj_body = {"a": 12}

        col = s.get_sync_collection("test_col")
        col2 = s.get_sync_collection("another_collection")
        col.put(obj_id, {"x": 13})
        col.put(obj_id, obj_body)
        assert col.find("a", 12) == [(obj_id, obj_body)]
        assert col.get(obj_id) == obj_body
        assert col2.get(obj_id) is None
        assert col.get("not exists") is None
        assert col.list() == [(obj_id, obj_body)]

        col.remove(obj_id)
        assert col.get(obj_id) is None

        s.stop()
        s.wait_completed(sync=True, timeout=5)


class TestMisc:
    """Various tests."""

    def test_invalid_col_name(self):
        """Test bad collection name raises exception."""
        s = Storage("sqlite://:memory:", threaded=True)
        s.start()
        try:
            with pytest.raises(ValueError, match="Invalid collection name:"):
                s.get_sync_collection("invalid%2345346?^$$$  /// : collection name")
        finally:
            s.stop()
            s.wait_completed(sync=True, timeout=10)

    def test_unsupoported_backend(self):
        """Test unsupported backed raises exception."""
        with pytest.raises(
            ValueError, match="Backend .* is not supported. Supported are"
        ):
            Storage._get_backend_instance("bad_back://test")


if __name__ == "__main__":
    pytest.main([os.path.basename(__file__)])



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the simple_data_request skill."""

import logging
from typing import cast
from unittest.mock import patch

import pytest

from aea.protocols.dialogue.base import DialogueMessage

from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.skills.simple_data_request.dialogues import HttpDialogues
from packages.fetchai.skills.simple_data_request.handlers import HttpHandler

from tests.test_packages.test_skills.test_simple_data_request.intermediate_class import (
    SimpleDataRequestTestCase,
)


class TestHttpHandler(SimpleDataRequestTestCase):
    """Test http handler of simple_data_request."""

    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.http_handler = cast(HttpHandler, cls._skill.skill_context.handlers.http)
        cls.logger = cls._skill.skill_context.logger

        cls.http_dialogues = cast(
            HttpDialogues, cls._skill.skill_context.http_dialogues
        )

        cls.data = b"some_body"

        cls.list_of_messages = (
            DialogueMessage(
                HttpMessage.Performative.REQUEST,
                {
                    "method": "some_method",
                    "url": "some_url",
                    "headers": "some_headers",
                    "version": "some_version",
                    "body": b"some_body",
                },
            ),
        )

    def test__init__(self):
        """Test the __init__ method of the http handler where ValueError is raise."""
        with pytest.raises(ValueError, match="No shared_state_key provided!"):
            self.http_handler.__init__(shared_state_key=None)

    def test_setup(self):
        """Test the setup method of the http handler."""
        assert self.http_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the http handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=HttpMessage.Performative.RESPONSE,
            method="some_method",
            url="some_url",
            headers="some_headers",
            version="some_version",
            body=b"some_body",
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid http message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_response(self):
        """Test the _handle_response method of the http handler."""
        # setup
        http_dialogue = self.prepare_skill_dialogue(
            dialogues=self.http_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=http_dialogue,
            performative=HttpMessage.Performative.RESPONSE,
            method="some_method",
            url="some_url",
            headers="some_headers",
            version="some_version",
            body=self.data,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"received http response={incoming_message} in dialogue={http_dialogue}.",
        )

        mock_logger.assert_any_call(
            logging.INFO, "updating shared_state with received data=b'some_body'!",
        )

        assert (
            self.skill.skill_context._agent_context.shared_state[
                self.mocked_shared_state_key
            ]
            == self.data
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the http handler."""
        # setup
        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            performative=HttpMessage.Performative.REQUEST,
            method="some_method",
            url="some_url",
            headers="some_headers",
            version="some_version",
            body=self.data,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle http message of performative={incoming_message.performative} in dialogue={self.http_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the http handler."""
        assert self.http_handler.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Module with generic utils of the aea cli."""
import os
from typing import Dict

import yaml
from click import ClickException

from aea.helpers.io import open_file


def load_yaml(filepath: str) -> Dict:
    """
    Read content from yaml file.

    :param filepath: str path to yaml file.

    :return: dict YAML content
    """
    with open_file(filepath, "r") as f:
        try:
            result = yaml.safe_load(f)
            return result if result is not None else {}
        except yaml.YAMLError as e:
            raise ClickException(
                "Loading yaml config from {} failed: {}".format(filepath, e)
            )


def is_readme_present(readme_path: str) -> bool:
    """
    Check is readme file present.

    This method is needed for proper testing.

    :param readme_path: path to readme file.

    :return: bool is readme file present.
    """
    return os.path.exists(readme_path)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the simple_data_request skill."""

import json
from typing import cast

import pytest

from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.skills.simple_data_request.behaviours import (
    HTTP_CLIENT_PUBLIC_ID,
    HttpRequestBehaviour,
)

from tests.test_packages.test_skills.test_simple_data_request.intermediate_class import (
    SimpleDataRequestTestCase,
)


class TestHttpRequestBehaviour(SimpleDataRequestTestCase):
    """Test http_request behaviour of http_request."""

    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.http_request_behaviour = cast(
            HttpRequestBehaviour, cls._skill.skill_context.behaviours.http_request
        )

    def test__init__i(self):
        """Test the __init__ method of the http_request behaviour."""
        assert self.http_request_behaviour.url == "some_url"
        assert self.http_request_behaviour.method == "some_method"
        assert self.http_request_behaviour.body == ""

    def test__init__ii(self):
        """Test the __init__ method of the http_request behaviour where ValueError is raise."""
        with pytest.raises(ValueError, match="Url, method and body must be provided."):
            self.http_request_behaviour.__init__(
                url=None, method="some_method", body="some_body"
            )

    def test_setup(self):
        """Test the setup method of the http_request behaviour."""
        assert self.http_request_behaviour.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_act(self):
        """Test the act method of the http_request behaviour."""
        # operation
        self.http_request_behaviour.act()

        # after
        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=HttpMessage,
            performative=HttpMessage.Performative.REQUEST,
            to=str(HTTP_CLIENT_PUBLIC_ID),
            sender=str(self.skill.skill_context.skill_id),
            method=self.mocked_method,
            url=self.mocked_url,
            headers="",
            version="",
            body=json.dumps(self.http_request_behaviour.body).encode("utf-8"),
        )
        assert has_attributes, error_str

    def test_teardown(self):
        """Test the teardown method of the http_request behaviour."""
        assert self.http_request_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the simple_service_registration skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.helpers.search.models import Description
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.simple_service_registration.behaviours import (
    ServiceRegistrationBehaviour,
)
from packages.fetchai.skills.simple_service_registration.strategy import Strategy

from tests.conftest import ROOT_DIR


class TestServiceRegistrationBehaviour(BaseSkillTestCase):
    """Test service behaviour of simple_service_registration."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "simple_service_registration"
    )
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.service_behaviour = cast(
            ServiceRegistrationBehaviour, cls._skill.skill_context.behaviours.service
        )
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)
        cls.logger = cls._skill.skill_context.logger

        cls.mocked_description_1 = Description({"foo1": 1, "bar1": 2})
        cls.mocked_description_2 = Description({"foo2": 1, "bar2": 2})

    def test_setup(self):
        """Test the setup method of the service behaviour."""
        # operation
        with patch.object(
            self.strategy,
            "get_location_description",
            return_value=self.mocked_description_1,
        ):
            with patch.object(
                self.strategy,
                "get_register_service_description",
                return_value=self.mocked_description_2,
            ):
                with patch.object(self.logger, "log") as mock_logger:
                    self.service_behaviour.setup()

        # after
        self.assert_quantity_in_outbox(2)

        # _register_agent
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=self.mocked_description_1,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "registering agent on SOEF.")

        # _register_service
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=self.mocked_description_2,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "registering service on SOEF.")

    def test_act(self):
        """Test the act method of the service behaviour."""
        assert self.service_behaviour.act() is None
        self.assert_quantity_in_outbox(0)

    def test_teardown(self):
        """Test the teardown method of the service behaviour."""
        # operation
        with patch.object(
            self.strategy,
            "get_location_description",
            return_value=self.mocked_description_1,
        ):
            with patch.object(
                self.strategy,
                "get_unregister_service_description",
                return_value=self.mocked_description_2,
            ):
                with patch.object(self.logger, "log") as mock_logger:
                    self.service_behaviour.teardown()

        # after
        self.assert_quantity_in_outbox(2)

        # _unregister_service
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=self.mocked_description_2,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "unregistering service from SOEF.")

        # _unregister_agent
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=self.mocked_description_1,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "unregistering agent from SOEF.")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the simple_data_request skill."""

import json
import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from vyper.utils import keccak256

from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.skills.fetch_beacon.dialogues import HttpDialogues
from packages.fetchai.skills.fetch_beacon.handlers import HttpHandler

from tests.conftest import ROOT_DIR


class TestHttpHandler(BaseSkillTestCase):
    """Test http handler of fetch_beacon skill."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "fetch_beacon")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls, **kwargs):
        """Setup the test class."""
        super().setup(**kwargs)
        cls.http_handler = cast(HttpHandler, cls._skill.skill_context.handlers.http)
        cls.logger = cls._skill.skill_context.logger

        cls.http_dialogues = cast(
            HttpDialogues, cls._skill.skill_context.http_dialogues
        )

        cls.list_of_messages = (
            DialogueMessage(
                HttpMessage.Performative.REQUEST,
                {
                    "method": "get",
                    "url": "some_url",
                    "headers": "",
                    "version": "",
                    "body": b"",
                },
            ),
        )

    def test_setup(self):
        """Test the setup method of the http handler."""
        assert self.http_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_response(self):
        """Test the _handle_response method of the http handler to a valid fetch beacon response."""
        # setup
        http_dialogue = self.prepare_skill_dialogue(
            dialogues=self.http_dialogues, messages=self.list_of_messages[:1],
        )

        test_response = {
            "result": {
                "block_id": {"hash": "00000000"},
                "block": {
                    "header": {
                        "height": "1",
                        "entropy": {"group_signature": "SIGNATURE"},
                    }
                },
            }
        }

        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=http_dialogue,
            performative=HttpMessage.Performative.RESPONSE,
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=json.dumps(test_response).encode("utf-8"),
        )

        # handle message
        self.http_handler.handle(incoming_message)

        # check that data was correctly entered into shared state
        beacon_data = {
            "entropy": keccak256("SIGNATURE".encode("utf-8")),
            "block_hash": bytes.fromhex("00000000"),
            "block_height": 1,
        }
        assert self.http_handler.context.shared_state["oracle_data"] == beacon_data

        # check that outbox is empty
        self.assert_quantity_in_outbox(0)

    def test_handle_response_invalid_body(self):
        """Test the _handle_response method of the http handler to an unexpected response."""
        # setup
        http_dialogue = self.prepare_skill_dialogue(
            dialogues=self.http_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=http_dialogue,
            performative=HttpMessage.Performative.RESPONSE,
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=b"{}",
        )

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        assert "oracle_data" not in self.http_handler.context.shared_state

        # after
        mock_logger.assert_any_call(
            logging.INFO, "entropy not present",
        )

    def test_handle__handle_unidentified_dialogue(self):
        """Test handling an unidentified dialogoue"""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=HttpMessage.Performative.RESPONSE,
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=b"{}",
        )

        # operation
        with patch.object(self.http_handler.context.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid message={incoming_message}, unidentified dialogue.",
        )

        self.assert_quantity_in_outbox(0)

    def test_teardown(self):
        """Test the teardown method of the http handler."""
        assert self.http_handler.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the misc utils for async code."""
import asyncio
import collections
import datetime
import logging
import time
from abc import ABC, abstractmethod
from asyncio import CancelledError
from asyncio.events import AbstractEventLoop, TimerHandle
from asyncio.futures import Future
from asyncio.tasks import FIRST_COMPLETED
from collections.abc import Iterable
from contextlib import contextmanager, suppress
from threading import Thread
from typing import (
    Any,
    Awaitable,
    Callable,
    Container,
    Generator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)


try:
    from asyncio import create_task  # pylint: disable=ungrouped-imports,unused-import
except ImportError:  # pragma: no cover
    # for python3.6!
    from asyncio import (  # type: ignore # noqa: F401 # pylint: disable=ungrouped-imports,unused-import
        ensure_future as create_task,
    )


_default_logger = logging.getLogger(__file__)


def ensure_list(value: Any) -> List:
    """Return [value] or list(value) if value is a sequence."""
    if isinstance(value, list):
        return value

    if isinstance(value, Iterable):
        return list(value)

    return [value]


not_set = object()


class AsyncState:
    """Awaitable state."""

    def __init__(
        self, initial_state: Any = None, states_enum: Optional[Container[Any]] = None
    ) -> None:
        """Init async state.

        :param initial_state: state to set on start.
        :param states_enum: container of valid states if not provided state not checked on set.
        """
        self._state = initial_state
        self._watchers: Set[Future] = set()
        self._callbacks: List[Callable[[Any], None]] = []
        self._states_enum = states_enum

    def set(self, state: Any) -> None:
        """Set state."""
        if self._states_enum is not None and state not in self._states_enum:
            raise ValueError(
                f"Unsupported state: {state}. Valid states are {self._states_enum}"
            )

        if self._state == state:  # pragma: no cover
            return

        self._state_changed(state)
        self._state = state

    def add_callback(self, callback_fn: Callable[[Any], None]) -> None:
        """
        Add callback to track state changes.

        :param callback_fn: callable object to be called on state changed.

        :return: None
        """
        self._callbacks.append(callback_fn)

    def get(self) -> Any:
        """Get state."""
        return self._state

    def _state_changed(self, state: Any) -> None:
        """Fulfill watchers for state."""
        for callback_fn in self._callbacks:
            try:
                callback_fn(state)
            except Exception:  # pylint: disable=broad-except
                _default_logger.exception(f"Exception on calling {callback_fn}")

        for watcher in list(self._watchers):
            if state not in watcher._states:  # type: ignore # pylint: disable=protected-access  # pragma: nocover
                continue
            if not watcher.done():
                watcher._loop.call_soon_threadsafe(  # pylint: disable=protected-access
                    self._watcher_result_callback(watcher), (self._state, state)
                )
            self._remove_watcher(watcher)

    def _remove_watcher(self, watcher: Future) -> None:
        """Remove watcher for state wait."""
        try:
            self._watchers.remove(watcher)
        except KeyError:
            pass

    @staticmethod
    def _watcher_result_callback(watcher: Future) -> Callable:
        """Create callback for watcher result."""
        # docstyle.
        def _callback(result: Any) -> None:
            if watcher.done():  # pragma: nocover
                return
            watcher.set_result(result)

        return _callback

    async def wait(self, state_or_states: Union[Any, Sequence[Any]]) -> Tuple[Any, Any]:
        """Wait state to be set.

        :param state_or_states: state or list of states.

        :return: tuple of previous state and new state.
        """
        states = ensure_list(state_or_states)

        if self._state in states:
            return (None, self._state)

        watcher: Future = Future()
        watcher._states = states  # type: ignore  # pylint: disable=protected-access
        self._watchers.add(watcher)
        try:
            return await watcher
        finally:
            self._remove_watcher(watcher)

    @contextmanager
    def transit(
        self, initial: Any = not_set, success: Any = not_set, fail: Any = not_set
    ) -> Generator:
        """
        Change state context according to success or not.

        :param initial: set state on context enter, not_set by default
        :param success: set state on context block done, not_set by default
        :param fail: set state on context block raises exception, not_set by default

        :return: None
        """
        try:
            if initial is not not_set:
                self.set(initial)
            yield
            if success is not not_set:
                self.set(success)
        except BaseException:
            if fail is not not_set:
                self.set(fail)
            raise


class PeriodicCaller:
    """
    Schedule a periodic call of callable using event loop.

    Used for periodic function run using asyncio.
    """

    def __init__(
        self,
        callback: Callable,
        period: float,
        start_at: Optional[datetime.datetime] = None,
        exception_callback: Optional[Callable[[Callable, Exception], None]] = None,
        loop: Optional[AbstractEventLoop] = None,
    ) -> None:
        """
        Init periodic caller.

        :param callback: function to call periodically
        :param period: period in seconds.
        :param start_at: optional first call datetime
        :param exception_callback: optional handler to call on exception raised.
        :param loop: optional asyncio event loop
        """
        self._loop = loop or asyncio.get_event_loop()
        self._periodic_callable = callback
        self._start_at = start_at or datetime.datetime.now()
        self._period = period
        self._timerhandle: Optional[TimerHandle] = None
        self._exception_callback = exception_callback

    def _callback(self) -> None:
        """Call on each scheduled call."""
        self._schedule_call()
        try:
            self._periodic_callable()
        except Exception as exception:  # pylint: disable=broad-except
            self.stop()
            if not self._exception_callback:  # pragma: nocover
                raise
            self._exception_callback(self._periodic_callable, exception)

    def _schedule_call(self) -> None:
        """Set schedule for call."""
        if self._timerhandle is None:
            ts = time.mktime(self._start_at.timetuple())
            delay = max(0, ts - time.time())
            self._timerhandle = self._loop.call_later(delay, self._callback)
        else:
            self._timerhandle = self._loop.call_later(self._period, self._callback)

    def start(self) -> None:
        """Activate period calls."""
        if self._timerhandle:  # pragma: nocover
            return

        self._schedule_call()

    def stop(self) -> None:
        """Remove from schedule."""
        if not self._timerhandle:  # pragma: nocover
            return

        self._timerhandle.cancel()
        self._timerhandle = None


class AnotherThreadTask:
    """
    Schedule a task to run on the loop in another thread.

    Provides better cancel behaviour: on cancel it will wait till cancelled completely.
    """

    def __init__(self, coro: Awaitable, loop: AbstractEventLoop) -> None:
        """
        Init the task.

        :param coro: coroutine to schedule
        :param loop: an event loop to schedule on.
        """
        self._loop = loop
        self._coro = coro
        self._task: Optional[asyncio.Task] = None
        self._future = asyncio.run_coroutine_threadsafe(self._get_task_result(), loop)

    async def _get_task_result(self) -> Any:
        """
        Get task result, should be run in target loop.

        :return: task result value or raise an exception if task failed
        """
        self._task = self._loop.create_task(self._coro)
        return await self._task

    def result(self, timeout: Optional[float] = None) -> Any:
        """
        Wait for coroutine execution result.

        :param timeout: optional timeout to wait in seconds.
        """
        return self._future.result(timeout)

    def cancel(self) -> None:
        """Cancel coroutine task execution in a target loop."""
        if self._task is None:
            self._loop.call_soon_threadsafe(self._future.cancel)
        else:
            self._loop.call_soon_threadsafe(self._task.cancel)

    def done(self) -> bool:
        """Check task is done."""
        return self._future.done()


class ThreadedAsyncRunner(Thread):
    """Util to run thread with event loop and execute coroutines inside."""

    def __init__(self, loop: Optional[AbstractEventLoop] = None) -> None:
        """
        Init threaded runner.

        :param loop: optional event loop. is it's running loop, threaded runner will use it.
        """
        self._loop = loop or asyncio.new_event_loop()
        if self._loop.is_closed():
            raise ValueError("Event loop closed.")  # pragma: nocover
        super().__init__(daemon=True)

    def start(self) -> None:
        """Start event loop in dedicated thread."""
        if self.is_alive() or self._loop.is_running():  # pragma: nocover
            return
        super().start()
        self.call(asyncio.sleep(0.001)).result(1)

    def run(self) -> None:
        """Run code inside thread."""
        _default_logger.debug("Starting threaded asyncio loop...")
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()
        _default_logger.debug("Asyncio loop has been stopped.")

    def call(self, coro: Awaitable) -> Any:
        """
        Run a coroutine inside the event loop.

        :param coro: a coroutine to run.
        """
        return AnotherThreadTask(coro, self._loop)

    def stop(self) -> None:
        """Stop event loop in thread."""
        _default_logger.debug("Stopping...")

        if not self.is_alive():  # pragma: nocover
            return

        if self._loop.is_running():
            _default_logger.debug("Stopping loop...")
            self._loop.call_soon_threadsafe(self._loop.stop)

        _default_logger.debug("Wait thread to join...")
        self.join(10)
        _default_logger.debug("Stopped.")


class ItemGetter:
    """Virtual queue like object to get items from getters function."""

    def __init__(self, getters: List[Callable]) -> None:
        """
        Init ItemGetter.

        :param getters: List of couroutines to be awaited.
        """
        if not getters:  # pragma: nocover
            raise ValueError("getters list can not be empty!")
        self._getters = getters
        self._items: collections.deque = collections.deque()

    async def get(self) -> Any:
        """Get item."""
        if not self._items:
            await self._wait()
        return self._items.pop()

    async def _wait(self) -> None:
        """Populate cache queue with items."""
        loop = asyncio.get_event_loop()
        try:
            tasks = [loop.create_task(getter()) for getter in self._getters]
            done, _ = await asyncio.wait(tasks, return_when=FIRST_COMPLETED)
            for task in done:
                self._items.append(await task)
        finally:
            for task in tasks:
                if task.done():
                    continue
                task.cancel()
                with suppress(CancelledError):
                    await task


class HandlerItemGetter(ItemGetter):
    """ItemGetter with handler passed."""

    @staticmethod
    def _make_getter(handler: Callable[[Any], None], getter: Callable) -> Callable:
        """
        Create getter for handler and item getter function.

        :param handler: callable with one position argument.
        :param getter: a couroutine to await for item

        :return: callable to return handler and item from getter.
        """
        # for pydocstyle
        async def _getter() -> Tuple[Callable, Any]:
            return handler, await getter()

        return _getter

    def __init__(self, getters: List[Tuple[Callable[[Any], None], Callable]]) -> None:
        """
        Init HandlerItemGetter.

        :param getters: List of tuples of handler and couroutine to be awaiteed for an item.
        """
        super().__init__(
            [self._make_getter(handler, getter) for handler, getter in getters]
        )


ready_future: Future = Future()
ready_future.set_result(None)


class Runnable(ABC):
    """
    Abstract Runnable class.

    Use to run async task in same event loop or in dedicated thread.
    Provides: start, stop sync methods to start and stop task
    Use wait_completed to await task was completed.
    """

    def __init__(
        self, loop: asyncio.AbstractEventLoop = None, threaded: bool = False
    ) -> None:
        """
        Init runnable.

        :param loop: asyncio event loop to use.
        :param threaded: bool. start in thread if True.

        :return: None
        """
        if loop and threaded:
            raise ValueError(
                "You can not set a loop in threaded mode. A dedicated loop will be created for each thread."
            )
        self._loop = loop
        self._threaded = threaded
        self._task: Optional[asyncio.Task] = None
        self._thread: Optional[Thread] = None
        self._completed_event: Optional[asyncio.Event] = None
        self._got_result = False
        self._was_cancelled = False
        self._is_running: bool = False
        self._stop_called = 0

    def start(self) -> bool:
        """
        Start runnable.

        :return: bool started or not.
        """
        if self._task and not self._task.done():
            _default_logger.debug(f"{self} already running")
            return False

        self._is_running = False
        self._got_result = False
        self._set_loop()
        self._completed_event = asyncio.Event(loop=self._loop)
        self._was_cancelled = False

        if self._stop_called > 0:
            # used in case of race when stop called before start!
            _default_logger.debug(f"{self} was already stopped before started!")
            self._stop_called = 0
            return True

        self._set_task()

        if self._threaded:
            self._thread = Thread(
                target=self._thread_target, name=self.__class__.__name__  # type: ignore # loop was set in set_loop
            )
            self._thread.setDaemon(True)
            self._thread.start()
        self._stop_called = 0
        return True

    def _thread_target(self) -> None:
        """Start event loop and task in the dedicated thread."""
        if not self._loop:
            raise ValueError("Call _set_loop() first!")  # pragma: nocover
        if not self._task:
            raise ValueError("Call _set_task() first!")  # pragma: nocover
        self._loop.run_until_complete(self._task)
        self._loop.stop()
        self._loop.close()

    def _set_loop(self) -> None:
        """Select and set loop."""
        if self._threaded:
            self._loop = asyncio.new_event_loop()
        else:
            try:
                self._loop = self._loop or asyncio.get_event_loop()
            except RuntimeError:
                self._loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self._loop)

    def _set_task(self) -> None:
        """Create task."""
        if not self._loop:  # pragma: nocover
            raise ValueError("Loop was not set.")
        self._task = self._loop.create_task(self._run_wrapper())
        _default_logger.debug(f"{self} task set")

    async def _run_wrapper(self) -> None:
        """Wrap run() method."""
        if not self._completed_event or not self._loop:  # pragma: nocover
            raise ValueError("Start was not called!")
        self._is_running = True
        try:
            with suppress(asyncio.CancelledError):
                return await self.run()
        finally:
            self._loop.call_soon_threadsafe(self._completed_event.set)
            self._is_running = False

    @property
    def is_running(self) -> bool:  # pragma: nocover
        """Get running state."""
        return self._is_running

    @abstractmethod
    async def run(self) -> Any:
        """Implement run logic respectfull to CancelError on termination."""

    def wait_completed(
        self, sync: bool = False, timeout: float = None, force_result: bool = False
    ) -> Awaitable:
        """
        Wait runnable execution completed.

        :param sync: bool. blocking wait
        :param timeout: float seconds
        :param force_result: check result even it was waited.

        :return: awaitable if sync is False, otherise None
        """
        if not self._task:
            _default_logger.warning("Runnable is not started")
            return ready_future

        if self._got_result and not force_result:
            return ready_future

        if sync:
            self._wait_sync(timeout)
            return ready_future

        return self._wait_async(timeout)

    def _wait_sync(self, timeout: Optional[float] = None) -> None:
        """Wait task completed in sync manner."""
        if self._task is None or not self._loop:  # pragma: nocover
            raise ValueError("task is not set!")

        if self._threaded or self._loop.is_running():
            start_time = time.time()

            while not self._task.done():
                time.sleep(0.01)
                if timeout is not None and time.time() - start_time > timeout:
                    raise asyncio.TimeoutError()

            if self._thread:
                self._thread.join(timeout)

            self._got_result = True
            if self._task.exception():
                raise self._task.exception()
        else:
            self._loop.run_until_complete(
                asyncio.wait_for(self._wait(), timeout=timeout)
            )

    def _wait_async(self, timeout: Optional[float] = None) -> Awaitable:
        if not self._threaded:
            return asyncio.wait_for(self._wait(), timeout=timeout)

        if self._task is None:  # pragma: nocover
            raise ValueError("task is not set!")

        # for threaded mode create a future and bind it to task
        loop = asyncio.get_event_loop()
        fut = loop.create_future()

        def done(task: Future) -> None:
            try:
                if fut.done():  # pragma: nocover
                    return
                if task.exception():
                    fut.set_exception(task.exception())
                else:  # pragma: nocover
                    fut.set_result(None)
            finally:
                self._got_result = True

        if self._task.done():
            done(self._task)
        else:
            self._task.add_done_callback(
                lambda task: loop.call_soon_threadsafe(lambda: done(task))
            )

        return fut

    async def _wait(self) -> None:
        """Wait internal method."""
        if not self._task or not self._completed_event:  # pragma: nocover
            raise ValueError("Not started")

        await self._completed_event.wait()

        try:
            await self._task
        finally:
            self._got_result = True

    def stop(self, force: bool = False) -> None:
        """Stop runnable."""
        _default_logger.debug(f"{self} is going to be stopped {self._task}")
        if not self._task or not self._loop:  # pragma: nocover
            self._stop_called += 1
            return

        if self._task.done():
            return

        self._loop.call_soon_threadsafe(self._task_cancel, force)

    def _task_cancel(self, force: bool = False) -> None:
        """Cancel task internal method."""
        if self._task is None:
            return

        if self._was_cancelled and not force:
            return

        self._was_cancelled = True
        self._task.cancel()

    def start_and_wait_completed(self, *args: Any, **kwargs: Any) -> Awaitable:
        """Alias for start and wait methods."""
        self.start()
        return self.wait_completed(*args, **kwargs)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the aea.configurations.validation module."""
from aea.configurations.validation import (
    SAME_MARK,
    filter_data,
    validate_data_with_pattern,
)


def test_compare_data_pattern():
    """Test validate_data_with_pattern."""
    errors = validate_data_with_pattern({"a": 12}, {"a": 13})
    assert not errors

    errors = validate_data_with_pattern({"a": 12}, {"a": "string"})
    assert errors
    assert (
        errors[0]
        == "For attribute `a` `str` data type is expected, but `int` was provided!"
    )

    errors = validate_data_with_pattern({"a": None}, {"a": int})
    assert not errors

    assert not validate_data_with_pattern(
        {"a": 12}, {"a": "${var}"}, skip_env_vars=True
    )
    assert not validate_data_with_pattern(
        {"a": "${var}"}, {"a": "string"}, skip_env_vars=True
    )

    errors = validate_data_with_pattern({"a": 12}, {"b": 12})
    assert errors
    assert errors[0] == "Attribute `a` is not allowed to be updated!"

    errors = validate_data_with_pattern({"a": {}}, {"a": {"b": 12}})
    assert errors
    assert errors[0] == "Attribute `a` is not allowed to be updated!"


def test_filter_data():
    """Test filter_data function."""
    assert filter_data(1, 1) == SAME_MARK
    assert filter_data(1, 2) == 2
    assert filter_data(2, 1) == 1

    assert filter_data({}, {}) == SAME_MARK
    assert filter_data({1: 2}, {1: 2}) == SAME_MARK
    assert filter_data({1: 2}, {1: 3}) == {1: 3}

    assert filter_data([1, 2, 3], [1, 2, 3]) == SAME_MARK
    assert filter_data([1, 2, 3], [1, 2]) == [1, 2]

    assert filter_data({1: {2: 3}, 3: {2: 1}}, {1: {2: 3}, 3: {2: 3}, 0: 0}) == {
        3: {2: 3},
        0: 0,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the helpers module of the tac negotiation."""

from pathlib import Path

from aea.helpers.search.models import (
    Attribute,
    Constraint,
    ConstraintType,
    DataModel,
    Description,
)
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.skills.tac_negotiation.helpers import (
    DEMAND_DATAMODEL_NAME,
    SUPPLY_DATAMODEL_NAME,
    _build_goods_datamodel,
    build_goods_description,
    build_goods_query,
)

from tests.conftest import ROOT_DIR


class TestHelpers(BaseSkillTestCase):
    """Test Helper module methods of tac control."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()

    def test_build_goods_datamodel_supply(self):
        """Test the _build_goods_datamodel of Helpers module for a supply."""
        good_ids = ["1", "2"]
        is_supply = True
        attributes = [
            Attribute("1", int, True, "A good on offer."),
            Attribute("2", int, True, "A good on offer."),
            Attribute("ledger_id", str, True, "The ledger for transacting."),
            Attribute(
                "currency_id",
                str,
                True,
                "The currency for pricing and transacting the goods.",
            ),
            Attribute("price", int, False, "The price of the goods in the currency."),
            Attribute(
                "fee",
                int,
                False,
                "The transaction fee payable by the buyer in the currency.",
            ),
            Attribute(
                "nonce", str, False, "The nonce to distinguish identical descriptions."
            ),
        ]
        expected_data_model = DataModel(SUPPLY_DATAMODEL_NAME, attributes)
        actual_data_model = _build_goods_datamodel(good_ids, is_supply)
        assert actual_data_model == expected_data_model

    def test_build_goods_datamodel_demand(self):
        """Test the _build_goods_datamodel of Helpers module for a demand."""
        good_ids = ["1", "2"]
        is_supply = False
        attributes = [
            Attribute("1", int, True, "A good on offer."),
            Attribute("2", int, True, "A good on offer."),
            Attribute("ledger_id", str, True, "The ledger for transacting."),
            Attribute(
                "currency_id",
                str,
                True,
                "The currency for pricing and transacting the goods.",
            ),
            Attribute("price", int, False, "The price of the goods in the currency."),
            Attribute(
                "fee",
                int,
                False,
                "The transaction fee payable by the buyer in the currency.",
            ),
            Attribute(
                "nonce", str, False, "The nonce to distinguish identical descriptions."
            ),
        ]
        expected_data_model = DataModel(DEMAND_DATAMODEL_NAME, attributes)
        actual_data_model = _build_goods_datamodel(good_ids, is_supply)
        assert actual_data_model == expected_data_model

    def test_build_goods_description_supply(self):
        """Test the build_goods_description of Helpers module for supply."""
        quantities_by_good_id = {"2": 5, "3": 10}
        currency_id = "1"
        ledger_id = "some_ledger_id"
        is_supply = True

        attributes = [
            Attribute("2", int, True, "A good on offer."),
            Attribute("3", int, True, "A good on offer."),
            Attribute("ledger_id", str, True, "The ledger for transacting."),
            Attribute(
                "currency_id",
                str,
                True,
                "The currency for pricing and transacting the goods.",
            ),
            Attribute("price", int, False, "The price of the goods in the currency."),
            Attribute(
                "fee",
                int,
                False,
                "The transaction fee payable by the buyer in the currency.",
            ),
            Attribute(
                "nonce", str, False, "The nonce to distinguish identical descriptions."
            ),
        ]
        expected_data_model = DataModel(SUPPLY_DATAMODEL_NAME, attributes)
        expected_values = {"currency_id": currency_id, "ledger_id": ledger_id}
        expected_values.update(quantities_by_good_id)
        expected_description = Description(expected_values, expected_data_model)

        actual_description = build_goods_description(
            quantities_by_good_id, currency_id, ledger_id, is_supply
        )
        assert actual_description == expected_description

    def test_build_goods_description_demand(self):
        """Test the build_goods_description of Helpers module for demand (same as above)."""
        quantities_by_good_id = {"2": 5, "3": 10}
        currency_id = "1"
        ledger_id = "some_ledger_id"
        is_supply = False

        attributes = [
            Attribute("2", int, True, "A good on offer."),
            Attribute("3", int, True, "A good on offer."),
            Attribute("ledger_id", str, True, "The ledger for transacting."),
            Attribute(
                "currency_id",
                str,
                True,
                "The currency for pricing and transacting the goods.",
            ),
            Attribute("price", int, False, "The price of the goods in the currency."),
            Attribute(
                "fee",
                int,
                False,
                "The transaction fee payable by the buyer in the currency.",
            ),
            Attribute(
                "nonce", str, False, "The nonce to distinguish identical descriptions."
            ),
        ]
        expected_data_model = DataModel(DEMAND_DATAMODEL_NAME, attributes)
        expected_values = {"currency_id": currency_id, "ledger_id": ledger_id}
        expected_values.update(quantities_by_good_id)
        expected_description = Description(expected_values, expected_data_model)

        actual_description = build_goods_description(
            quantities_by_good_id, currency_id, ledger_id, is_supply
        )
        assert actual_description == expected_description

    def test_build_goods_query(self):
        """Test the build_goods_query of Helpers module."""
        good_ids = ["2", "3"]
        currency_id = "1"
        ledger_id = "some_ledger_id"
        is_searching_for_sellers = True

        attributes = [
            Attribute("2", int, True, "A good on offer."),
            Attribute("3", int, True, "A good on offer."),
            Attribute("ledger_id", str, True, "The ledger for transacting."),
            Attribute(
                "currency_id",
                str,
                True,
                "The currency for pricing and transacting the goods.",
            ),
            Attribute("price", int, False, "The price of the goods in the currency."),
            Attribute(
                "fee",
                int,
                False,
                "The transaction fee payable by the buyer in the currency.",
            ),
            Attribute(
                "nonce", str, False, "The nonce to distinguish identical descriptions."
            ),
        ]
        expected_data_model = DataModel(SUPPLY_DATAMODEL_NAME, attributes)

        expected_constraints = [
            Constraint("2", ConstraintType(">=", 1)),
            Constraint("3", ConstraintType(">=", 1)),
            Constraint("ledger_id", ConstraintType("==", ledger_id)),
            Constraint("currency_id", ConstraintType("==", currency_id)),
        ]

        actual_query = build_goods_query(
            good_ids, currency_id, ledger_id, is_searching_for_sellers
        )

        constraints = [
            (c.constraint_type.type, c.constraint_type.value)
            for c in actual_query.constraints[0].constraints
        ]
        for constraint in expected_constraints:
            assert (
                constraint.constraint_type.type,
                constraint.constraint_type.value,
            ) in constraints
        assert actual_query.model == expected_data_model

    def test_build_goods_query_1_good(self):
        """Test the build_goods_query of Helpers module where there is 1 good."""
        good_ids = ["2"]
        currency_id = "1"
        ledger_id = "some_ledger_id"
        is_searching_for_sellers = True

        attributes = [
            Attribute("2", int, True, "A good on offer."),
            Attribute("ledger_id", str, True, "The ledger for transacting."),
            Attribute(
                "currency_id",
                str,
                True,
                "The currency for pricing and transacting the goods.",
            ),
            Attribute("price", int, False, "The price of the goods in the currency."),
            Attribute(
                "fee",
                int,
                False,
                "The transaction fee payable by the buyer in the currency.",
            ),
            Attribute(
                "nonce", str, False, "The nonce to distinguish identical descriptions."
            ),
        ]
        expected_data_model = DataModel(SUPPLY_DATAMODEL_NAME, attributes)

        expected_constraints = [
            Constraint("2", ConstraintType(">=", 1)),
            Constraint("ledger_id", ConstraintType("==", ledger_id)),
            Constraint("currency_id", ConstraintType("==", currency_id)),
        ]

        actual_query = build_goods_query(
            good_ids, currency_id, ledger_id, is_searching_for_sellers
        )

        for constraint in expected_constraints:
            assert constraint in actual_query.constraints
        assert actual_query.model == expected_data_model



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the base classes for the skills."""
import datetime
import inspect
import logging
import queue
import re
import types
from abc import ABC, abstractmethod
from copy import copy
from logging import Logger
from pathlib import Path
from queue import Queue
from types import SimpleNamespace
from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union, cast

from aea.common import Address
from aea.components.base import Component, load_aea_package
from aea.configurations.base import (
    ComponentType,
    PublicId,
    SkillComponentConfiguration,
    SkillConfig,
)
from aea.configurations.loader import load_component_configuration
from aea.context.base import AgentContext
from aea.exceptions import (
    AEAActException,
    AEAComponentLoadException,
    AEAHandleException,
    AEAInstantiationException,
    _StopRuntime,
    enforce,
    parse_exception,
)
from aea.helpers.base import _get_aea_logger_name_prefix, load_module
from aea.helpers.logging import AgentLoggerAdapter
from aea.helpers.storage.generic_storage import Storage
from aea.mail.base import Envelope, EnvelopeContext
from aea.multiplexer import MultiplexerStatus, OutBox
from aea.protocols.base import Message
from aea.skills.tasks import TaskManager


_default_logger = logging.getLogger(__name__)


class SkillContext:
    """This class implements the context of a skill."""

    def __init__(
        self,
        agent_context: Optional[AgentContext] = None,
        skill: Optional["Skill"] = None,
    ) -> None:
        """
        Initialize a skill context.

        :agent_context: the agent context.
        :skill: the skill.
        """
        self._agent_context = agent_context  # type: Optional[AgentContext]
        self._in_queue = Queue()  # type: Queue
        self._skill = skill  # type: Optional[Skill]

        self._is_active = True  # type: bool
        self._new_behaviours_queue = queue.Queue()  # type: Queue
        self._new_handlers_queue = queue.Queue()  # type: Queue
        self._logger: Optional[Logger] = None

    @property
    def logger(self) -> Logger:
        """Get the logger."""
        if self._logger is None:
            return _default_logger
        return self._logger

    @logger.setter
    def logger(self, logger_: Logger) -> None:
        """Set the logger."""
        self._logger = logger_

    def _get_agent_context(self) -> AgentContext:
        """Get the agent context."""
        if self._agent_context is None:  # pragma: nocover
            raise ValueError("Agent context not set yet.")
        return self._agent_context

    def set_agent_context(self, agent_context: AgentContext) -> None:
        """Set the agent context."""
        self._agent_context = agent_context

    @property
    def shared_state(self) -> Dict[str, Any]:
        """Get the shared state dictionary."""
        return self._get_agent_context().shared_state

    @property
    def agent_name(self) -> str:
        """Get agent name."""
        return self._get_agent_context().agent_name

    @property
    def skill_id(self) -> PublicId:
        """Get the skill id of the skill context."""
        if self._skill is None:
            raise ValueError("Skill not set yet.")  # pragma: nocover
        return self._skill.configuration.public_id

    @property
    def is_active(self) -> bool:
        """Get the status of the skill (active/not active)."""
        return self._is_active

    @is_active.setter
    def is_active(self, value: bool) -> None:
        """Set the status of the skill (active/not active)."""
        self._is_active = value
        self.logger.debug(
            "New status of skill {}: is_active={}".format(
                self.skill_id, self._is_active
            )
        )

    @property
    def new_behaviours(self) -> "Queue[Behaviour]":
        """
        Queue for the new behaviours.

        This queue can be used to send messages to the framework
        to request the registration of a behaviour.

        :return the queue of new behaviours.
        """
        return self._new_behaviours_queue

    @property
    def new_handlers(self) -> "Queue[Handler]":
        """
        Queue for the new handlers.

        This queue can be used to send messages to the framework
        to request the registration of a handler.

        :return the queue of new handlers.
        """
        return self._new_handlers_queue

    @property
    def agent_addresses(self) -> Dict[str, str]:
        """Get addresses."""
        return self._get_agent_context().addresses

    @property
    def agent_address(self) -> str:
        """Get address."""
        return self._get_agent_context().address

    @property
    def connection_status(self) -> MultiplexerStatus:
        """Get connection status."""
        return self._get_agent_context().connection_status

    @property
    def outbox(self) -> OutBox:
        """Get outbox."""
        return self._get_agent_context().outbox

    @property
    def storage(self) -> Optional[Storage]:
        """Get optional storage for agent."""
        return self._get_agent_context().storage

    @property
    def message_in_queue(self) -> Queue:
        """Get message in queue."""
        return self._in_queue

    @property
    def decision_maker_message_queue(self) -> Queue:
        """Get message queue of decision maker."""
        return self._get_agent_context().decision_maker_message_queue

    @property
    def decision_maker_handler_context(self) -> SimpleNamespace:
        """Get decision maker handler context."""
        return cast(
            SimpleNamespace, self._get_agent_context().decision_maker_handler_context
        )

    @property
    def task_manager(self) -> TaskManager:
        """Get behaviours of the skill."""
        if self._skill is None:
            raise ValueError("Skill not initialized.")
        return self._get_agent_context().task_manager

    @property
    def default_ledger_id(self) -> str:
        """Get the default ledger id."""
        return self._get_agent_context().default_ledger_id

    @property
    def currency_denominations(self) -> Dict[str, str]:
        """Get a dictionary mapping ledger ids to currency denominations."""
        return self._get_agent_context().currency_denominations

    @property
    def search_service_address(self) -> Address:
        """Get the address of the search service."""
        return self._get_agent_context().search_service_address

    @property
    def decision_maker_address(self) -> Address:
        """Get the address of the decision maker."""
        return self._get_agent_context().decision_maker_address

    @property
    def handlers(self) -> SimpleNamespace:
        """Get handlers of the skill."""
        if self._skill is None:
            raise ValueError("Skill not initialized.")
        return SimpleNamespace(**self._skill.handlers)

    @property
    def behaviours(self) -> SimpleNamespace:
        """Get behaviours of the skill."""
        if self._skill is None:
            raise ValueError("Skill not initialized.")
        return SimpleNamespace(**self._skill.behaviours)

    @property
    def namespace(self) -> SimpleNamespace:
        """Get the agent context namespace."""
        return self._get_agent_context().namespace

    def __getattr__(self, item: Any) -> Any:
        """Get attribute."""
        return super().__getattribute__(item)  # pragma: no cover

    def send_to_skill(
        self,
        message_or_envelope: Union[Message, Envelope],
        context: Optional[EnvelopeContext] = None,
    ) -> None:
        """
        Send message or envelope to another skill.

        :param message_or_envelope: envelope to send to another skill.
        if message passed it will be wrapped into envelope with optional envelope context.

        :return: None
        """
        if self._agent_context is None:  # pragma: nocover
            raise ValueError("agent context was not set!")
        self._agent_context.send_to_skill(message_or_envelope, context)


class SkillComponent(ABC):
    """This class defines an abstract interface for skill component classes."""

    def __init__(
        self,
        name: str,
        skill_context: SkillContext,
        configuration: Optional[SkillComponentConfiguration] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a skill component.

        :param name: the name of the component.
        :param configuration: the configuration for the component.
        :param skill_context: the skill context.
        """
        if name is None:
            raise ValueError("SkillComponent name is not provided.")
        if skill_context is None:
            raise ValueError("SkillConext is not provided")
        if configuration is None:
            class_name = type(self).__name__
            configuration = SkillComponentConfiguration(class_name=class_name, **kwargs)
        self._configuration = configuration
        self._name = name
        self._context = skill_context
        if len(kwargs) != 0:
            self.context.logger.warning(
                "The kwargs={} passed to {} have not been set!".format(kwargs, name)
            )

    @property
    def name(self) -> str:
        """Get the name of the skill component."""
        return self._name

    @property
    def context(self) -> SkillContext:
        """Get the context of the skill component."""
        return self._context

    @property
    def skill_id(self) -> PublicId:
        """Get the skill id of the skill component."""
        return self.context.skill_id

    @property
    def configuration(self) -> SkillComponentConfiguration:
        """Get the skill component configuration."""
        if self._configuration is None:
            raise ValueError("Configuration not set.")  # pragma: nocover
        return self._configuration

    @property
    def config(self) -> Dict[Any, Any]:
        """Get the config of the skill component."""
        return self.configuration.args

    @abstractmethod
    def setup(self) -> None:
        """
        Implement the setup.

        :return: None
        """
        super_obj = super()
        if hasattr(super_obj, "setup"):
            super_obj.setup()  # type: ignore  # pylint: disable=no-member

    @abstractmethod
    def teardown(self) -> None:
        """
        Implement the teardown.

        :return: None
        """
        super_obj = super()
        if hasattr(super_obj, "teardown"):
            super_obj.teardown()  # type: ignore  # pylint: disable=no-member

    @classmethod
    @abstractmethod
    def parse_module(
        cls,
        path: str,
        configs: Dict[str, SkillComponentConfiguration],
        skill_context: SkillContext,
    ) -> dict:
        """Parse the component module."""


class AbstractBehaviour(SkillComponent, ABC):
    """
    Abstract behaviour for periodical calls.

    tick_interval: float, interval to call behaviour's act.
    start_at: optional datetime, when to start periodical calls.
    """

    _tick_interval: float = 0.001
    _start_at: Optional[datetime.datetime] = None

    @property
    def tick_interval(self) -> float:
        """Get the tick_interval in seconds."""
        return self._tick_interval

    @property
    def start_at(self) -> Optional[datetime.datetime]:
        """Get the start time of the behaviour."""
        return self._start_at


class Behaviour(AbstractBehaviour, ABC):
    """This class implements an abstract behaviour."""

    @abstractmethod
    def act(self) -> None:
        """
        Implement the behaviour.

        :return: None
        """

    def is_done(self) -> bool:  # pylint: disable=no-self-use
        """Return True if the behaviour is terminated, False otherwise."""
        return False

    def act_wrapper(self) -> None:
        """Wrap the call of the action. This method must be called only by the framework."""
        try:
            self.act()
        except _StopRuntime:
            raise
        except Exception as e:  # pylint: disable=broad-except
            e_str = parse_exception(e)
            raise AEAActException(
                f"An error occured during act of behaviour {self.context.skill_id}/{type(self).__name__}:\n{e_str}"
            )

    @classmethod
    def parse_module(  # pylint: disable=arguments-differ
        cls,
        path: str,
        behaviour_configs: Dict[str, SkillComponentConfiguration],
        skill_context: SkillContext,
    ) -> Dict[str, "Behaviour"]:
        """
        Parse the behaviours module.

        :param path: path to the Python module containing the Behaviour classes.
        :param behaviour_configs: a list of behaviour configurations.
        :param skill_context: the skill context
        :return: a list of Behaviour.
        """
        return _parse_module(path, behaviour_configs, skill_context, Behaviour)


class Handler(SkillComponent, ABC):
    """This class implements an abstract behaviour."""

    SUPPORTED_PROTOCOL = None  # type: Optional[PublicId]

    @abstractmethod
    def handle(self, message: Message) -> None:
        """
        Implement the reaction to a message.

        :param message: the message
        :return: None
        """

    def handle_wrapper(self, message: Message) -> None:
        """Wrap the call of the handler. This method must be called only by the framework."""
        try:
            self.handle(message)
        except _StopRuntime:
            raise
        except Exception as e:  # pylint: disable=broad-except
            e_str = parse_exception(e)
            raise AEAHandleException(
                f"An error occured during handle of handler {self.context.skill_id}/{type(self).__name__}:\n{e_str}"
            )

    @classmethod
    def parse_module(  # pylint: disable=arguments-differ
        cls,
        path: str,
        handler_configs: Dict[str, SkillComponentConfiguration],
        skill_context: SkillContext,
    ) -> Dict[str, "Handler"]:
        """
        Parse the handler module.

        :param path: path to the Python module containing the Handler class.
        :param handler_configs: the list of handler configurations.
        :param skill_context: the skill context
        :return: an handler, or None if the parsing fails.
        """
        return _parse_module(path, handler_configs, skill_context, Handler)


class Model(SkillComponent, ABC):
    """This class implements an abstract model."""

    def __init__(
        self,
        name: str,
        skill_context: SkillContext,
        configuration: Optional[SkillComponentConfiguration] = None,
        keep_terminal_state_dialogues: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a model.

        :param name: the name of the component.
        :param configuration: the configuration for the component.
        :param skill_context: the skill context.
        :param keep_terminal_state_dialogues: specify do dialogues in terminal state should stay or not

        :return: None
        """
        super().__init__(name, skill_context, configuration=configuration, **kwargs)

        # used by dialogues if mixed with the Model
        if keep_terminal_state_dialogues is not None:
            self._keep_terminal_state_dialogues = keep_terminal_state_dialogues

    def setup(self) -> None:
        """Set the class up."""
        super_obj = super()
        if hasattr(super_obj, "setup"):
            super_obj.setup()  # type: ignore  # pylint: disable=no-member

    def teardown(self) -> None:
        """Tear the class down."""
        super_obj = super()
        if hasattr(super_obj, "teardown"):
            super_obj.teardown()  # type: ignore  # pylint: disable=no-member

    @classmethod
    def parse_module(  # pylint: disable=arguments-differ
        cls,
        path: str,
        model_configs: Dict[str, SkillComponentConfiguration],
        skill_context: SkillContext,
    ) -> Dict[str, "Model"]:
        """
        Parse the model module.

        :param path: path to the Python skill module.
        :param model_configs: a list of model configurations.
        :param skill_context: the skill context
        :return: a list of Model.
        """
        return _parse_module(path, model_configs, skill_context, Model)


class Skill(Component):
    """This class implements a skill."""

    __slots__ = ("_skill_context", "_handlers", "_behaviours", "_models")

    def __init__(
        self,
        configuration: SkillConfig,
        skill_context: Optional[SkillContext] = None,
        handlers: Optional[Dict[str, Handler]] = None,
        behaviours: Optional[Dict[str, Behaviour]] = None,
        models: Optional[Dict[str, Model]] = None,
        **kwargs: Any,
    ):
        """
        Initialize a skill.

        :param configuration: the skill configuration.
        :param skill_context: the skill context.
        :param handlers: dictionary of handlers.
        :param behaviours: dictionary of behaviours.
        :param models: dictionary of models.
        """
        if kwargs is not None:
            pass
        super().__init__(configuration)
        self._skill_context = (
            skill_context if skill_context is not None else SkillContext()
        )
        self._handlers = (
            {} if handlers is None else handlers
        )  # type: Dict[str, Handler]
        self._behaviours = (
            {} if behaviours is None else behaviours
        )  # type: Dict[str, Behaviour]
        self._models = {} if models is None else models  # type: Dict[str, Model]

        self._skill_context._skill = self
        self._set_models_on_context()

    def _set_models_on_context(self) -> None:
        """Set the models on the skill context."""
        if self._models != {}:
            for model_id, model_instance in self._models.items():
                if getattr(self._skill_context, model_id, None) is None:
                    setattr(self._skill_context, model_id, model_instance)

    @property
    def skill_context(self) -> SkillContext:
        """Get the skill context."""
        if self._skill_context is None:
            raise ValueError("Skill context not set.")  # pragma: nocover
        return self._skill_context

    @property
    def handlers(self) -> Dict[str, Handler]:
        """Get the handlers."""
        return self._handlers

    @property
    def behaviours(self) -> Dict[str, Behaviour]:
        """Get the handlers."""
        return self._behaviours

    @property
    def models(self) -> Dict[str, Model]:
        """Get the handlers."""
        return self._models

    @classmethod
    def from_dir(
        cls, directory: str, agent_context: AgentContext, **kwargs: Any
    ) -> "Skill":
        """
        Load the skill from a directory.

        :param directory: the directory to the skill package.
        :param agent_context: the skill context
        :return: the skill object.
        """
        configuration = cast(
            SkillConfig,
            load_component_configuration(ComponentType.SKILL, Path(directory)),
        )
        configuration.directory = Path(directory)
        return Skill.from_config(configuration, agent_context, **kwargs)

    @property
    def logger(self) -> Logger:
        """
        Get the logger.

        In the case of a skill, return the
        logger provided by the skill context.
        """
        return self.skill_context.logger

    @logger.setter
    def logger(self, *args: str) -> None:
        """Set the logger."""
        raise ValueError("Cannot set logger to a skill component.")

    @classmethod
    def from_config(
        cls, configuration: SkillConfig, agent_context: AgentContext, **kwargs: Any
    ) -> "Skill":
        """
        Load the skill from configuration.

        :param configuration: a skill configuration. Must be associated with a directory.
        :param agent_context: the agent context.
        :return: the skill.
        """

        if configuration.directory is None:  # pragma: nocover
            raise ValueError("Configuration must be associated with a directory.")

        # we put the initialization here because some skill components
        # might need some info from the skill
        # (e.g. see https://github.com/fetchai/agents-aea/issues/1095)
        skill_context = SkillContext()
        skill_context.set_agent_context(agent_context)
        logger_name = f"aea.packages.{configuration.author}.skills.{configuration.name}"
        logger_name = _get_aea_logger_name_prefix(logger_name, agent_context.agent_name)
        _logger = AgentLoggerAdapter(
            logging.getLogger(logger_name), agent_context.agent_name
        )
        skill_context.logger = cast(Logger, _logger)

        skill_component_loader = _SkillComponentLoader(
            configuration, skill_context, **kwargs
        )
        skill = skill_component_loader.load_skill()
        return skill


def _parse_module(
    path: str,
    component_configs: Dict[str, SkillComponentConfiguration],
    skill_context: SkillContext,
    component_class: Type,
) -> Dict[str, Any]:
    """
    Parse a module to find skill component classes, and instantiate them.

    This is a private framework function,
     used in SkillComponentClass.parse_module.

    :param path: path to the Python module.
    :param component_configs: the component configurations.
    :param skill_context: the skill context.
    :param component_class: the class of the skill components to be loaded.
    :return: A mapping from skill component name to the skill component instance.
    """
    components: Dict[str, Any] = {}
    component_type_name = component_class.__name__.lower()
    component_type_name_plural = component_type_name + "s"
    if component_configs == {}:
        return components
    component_names = set(config.class_name for _, config in component_configs.items())
    component_module = load_module(component_type_name_plural, Path(path))
    classes = inspect.getmembers(component_module, inspect.isclass)
    component_classes = list(
        filter(
            lambda x: any(re.match(component, x[0]) for component in component_names)
            and issubclass(x[1], component_class)
            and not str.startswith(x[1].__module__, "aea.")
            and not str.startswith(
                x[1].__module__,
                f"packages.{skill_context.skill_id.author}.skills.{skill_context.skill_id.name}",
            ),
            classes,
        )
    )

    name_to_class = dict(component_classes)
    _print_warning_message_for_non_declared_skill_components(
        skill_context,
        set(name_to_class.keys()),
        {
            component_config.class_name
            for component_config in component_configs.values()
        },
        component_type_name_plural,
        path,
    )
    for component_id, component_config in component_configs.items():
        component_class_name = cast(str, component_config.class_name)
        skill_context.logger.debug(
            f"Processing {component_type_name} {component_class_name}"
        )
        if not component_id.isidentifier():
            raise AEAComponentLoadException(  # pragma: nocover
                f"'{component_id}' is not a valid identifier."
            )
        component_class = name_to_class.get(component_class_name, None)
        if component_class is None:
            skill_context.logger.warning(
                f"{component_type_name.capitalize()} '{component_class_name}' cannot be found."
            )
        else:
            try:
                component = component_class(
                    name=component_id,
                    configuration=component_config,
                    skill_context=skill_context,
                    **dict(component_config.args),
                )
            except Exception as e:  # pylint: disable=broad-except # pragma: nocover
                e_str = parse_exception(e)
                raise AEAInstantiationException(
                    f"An error occured during instantiation of component {skill_context.skill_id}/{component_config.class_name}:\n{e_str}"
                )
            components[component_id] = component

    return components


def _print_warning_message_for_non_declared_skill_components(
    skill_context: SkillContext,
    classes: Set[str],
    config_components: Set[str],
    item_type: str,
    skill_path: str,
) -> None:
    """Print a warning message if a skill component is not declared in the config files."""
    for class_name in classes.difference(config_components):
        skill_context.logger.warning(
            "Class {} of type {} found but not declared in the configuration file {}.".format(
                class_name, item_type, skill_path
            )
        )


_SKILL_COMPONENT_TYPES = Type[Union[Handler, Behaviour, Model]]

_ComponentsHelperIndex = Dict[_SKILL_COMPONENT_TYPES, Dict[str, SkillComponent]]
"""
Helper index to store component instances.
"""


class _SkillComponentLoadingItem:  # pylint: disable=too-few-public-methods
    """Class to represent a triple (component name, component configuration, component class)."""

    def __init__(
        self,
        name: str,
        config: SkillComponentConfiguration,
        class_: Type[SkillComponent],
        type_: _SKILL_COMPONENT_TYPES,
    ):
        """Initialize the item."""
        self.name = name
        self.config = config
        self.class_ = class_
        self.type_ = type_


class _SkillComponentLoader:
    """This class implements the loading policy for skill components."""

    def __init__(
        self, configuration: SkillConfig, skill_context: SkillContext, **kwargs: Any
    ):
        """Initialize the helper class."""
        enforce(
            configuration.directory is not None,
            "Configuration not associated to directory.",
        )
        self.configuration = configuration
        self.skill_directory = cast(Path, configuration.directory)
        self.skill_context = skill_context
        self.kwargs = kwargs

        self._all_component_names: Set[str] = set()

        self.skill = Skill(self.configuration, self.skill_context, **self.kwargs)
        self.skill_dotted_path = f"packages.{self.configuration.public_id.author}.skills.{self.configuration.public_id.name}"

    def load_skill(self) -> Skill:
        """Load the skill."""
        load_aea_package(self.configuration)
        python_modules: Set[Path] = self._get_python_modules()
        declared_component_classes: Dict[
            _SKILL_COMPONENT_TYPES, Dict[str, SkillComponentConfiguration]
        ] = self._get_declared_skill_component_configurations()
        self._all_component_names = {
            config.class_name
            for _, types_ in declared_component_classes.items()
            for _, config in types_.items()
        }
        component_classes_by_path: Dict[
            Path, Set[Tuple[str, Type[SkillComponent]]]
        ] = self._load_component_classes(python_modules)
        component_loading_items = self._match_class_and_configurations(
            component_classes_by_path, declared_component_classes
        )
        components = self._get_component_instances(component_loading_items)
        self._update_skill(components)
        return self.skill

    def _update_skill(self, components: _ComponentsHelperIndex) -> None:
        self.skill.handlers.update(
            cast(Dict[str, Handler], components.get(Handler, {}))
        )
        self.skill.behaviours.update(
            cast(Dict[str, Behaviour], components.get(Behaviour, {}))
        )
        self.skill.models.update(cast(Dict[str, Model], components.get(Model, {})))
        self.skill._set_models_on_context()  # pylint: disable=protected-access

    def _get_python_modules(self) -> Set[Path]:
        """
        Get all the Python modules of the skill package.

        We ignore '__pycache__' Python modules as they are not relevant.

        :return: a set of paths pointing to all the Python modules in the skill.
        """
        ignore_regex = "__pycache__*"
        all_python_modules = self.skill_directory.rglob("*.py")
        module_paths: Set[Path] = set(
            map(
                lambda p: Path(p).relative_to(self.skill_directory),
                filter(
                    lambda x: not re.match(ignore_regex, x.name), all_python_modules
                ),
            )
        )
        return module_paths

    @classmethod
    def _compute_module_dotted_path(cls, module_path: Path) -> str:
        """Compute the dotted path for a skill module."""
        suffix = ".".join(module_path.with_name(module_path.stem).parts)
        return suffix

    def _filter_classes(
        self, classes: List[Tuple[str, Type]]
    ) -> List[Tuple[str, Type[SkillComponent]]]:
        """
        Filter classes of skill components.

        :param classes: a list of pairs (class name, class object)
        :return: a list of the same kind, but filtered with only skill component classes.
        """
        filtered_classes = filter(
            lambda name_and_class: any(
                re.match(component, name_and_class[0])
                for component in self._all_component_names
            )
            and issubclass(name_and_class[1], SkillComponent)
            and not str.startswith(name_and_class[1].__module__, "aea.")
            and not str.startswith(
                name_and_class[1].__module__, self.skill_dotted_path
            ),
            classes,
        )
        classes = list(filtered_classes)
        return cast(List[Tuple[str, Type[SkillComponent]]], classes)

    def _load_component_classes(
        self, module_paths: Set[Path]
    ) -> Dict[Path, Set[Tuple[str, Type[SkillComponent]]]]:
        """
        Load component classes from Python modules.

        :param module_paths: a set of paths to Python modules.
        :return: a mapping from path to skill component classes in that module
          (containing potential duplicates). Skill components in one path
          are
        """
        module_to_classes: Dict[Path, Set[Tuple[str, Type[SkillComponent]]]] = {}
        for module_path in module_paths:
            self.skill_context.logger.debug(f"Trying to load module {module_path}")
            module_dotted_path: str = self._compute_module_dotted_path(module_path)
            component_module: types.ModuleType = load_module(
                module_dotted_path, self.skill_directory / module_path
            )
            classes: List[Tuple[str, Type]] = inspect.getmembers(
                component_module, inspect.isclass
            )
            filtered_classes: List[
                Tuple[str, Type[SkillComponent]]
            ] = self._filter_classes(classes)
            module_to_classes[module_path] = set(filtered_classes)
        return module_to_classes

    def _get_declared_skill_component_configurations(
        self,
    ) -> Dict[_SKILL_COMPONENT_TYPES, Dict[str, SkillComponentConfiguration]]:
        """
        Get all the declared skill component configurations.

        :return:
        """
        handlers_by_id = dict(self.configuration.handlers.read_all())
        behaviours_by_id = dict(self.configuration.behaviours.read_all())
        models_by_id = dict(self.configuration.models.read_all())

        result: Dict[
            _SKILL_COMPONENT_TYPES, Dict[str, SkillComponentConfiguration]
        ] = {}
        for component_type, components_by_id in [
            (Handler, handlers_by_id),
            (Behaviour, behaviours_by_id),
            (Model, models_by_id),
        ]:
            for component_id, component_config in components_by_id.items():
                result.setdefault(component_type, {})[component_id] = component_config  # type: ignore
        return result

    def _get_component_instances(
        self, component_loading_items: List[_SkillComponentLoadingItem],
    ) -> _ComponentsHelperIndex:
        """
        Instantiate classes declared in configuration files.

        :param component_loading_items: a list of loading items.
        :return: the instances of the skill components.
        """
        result: _ComponentsHelperIndex = {}
        for item in component_loading_items:
            instance = item.class_(
                name=item.name,
                configuration=item.config,
                skill_context=self.skill_context,
                **item.config.args,
            )
            result.setdefault(item.type_, {})[item.name] = instance
        return result

    @classmethod
    def _get_skill_component_type(
        cls, skill_component_type: Type[SkillComponent],
    ) -> Type[Union[Handler, Behaviour, Model]]:
        """Get the concrete skill component type."""
        parent_skill_component_types = list(
            filter(
                lambda class_: class_ in (Handler, Behaviour, Model),
                skill_component_type.__mro__,
            )
        )
        enforce(
            len(parent_skill_component_types) == 1,
            f"Class {skill_component_type.__name__} in module {skill_component_type.__module__} is not allowed to inherit from more than one skill component type. Found: {parent_skill_component_types}.",
        )
        return cast(
            Type[Union[Handler, Behaviour, Model]], parent_skill_component_types[0]
        )

    def _match_class_and_configurations(
        self,
        component_classes_by_path: Dict[Path, Set[Tuple[str, Type[SkillComponent]]]],
        declared_component_classes: Dict[
            _SKILL_COMPONENT_TYPES, Dict[str, SkillComponentConfiguration]
        ],
    ) -> List[_SkillComponentLoadingItem]:
        """
        Match skill component classes to their configurations.

        Given a class of a skill component, we can disambiguate it in three ways:
        - by its name
        - by its type (one of 'Handler', 'Behaviour', 'Model')
        - whether the user has set the 'file_path' field.
        If one of the skill component cannot be disambiguated, we raise error.

        In this function, the above criteria are applied in that order.

        :param component_classes_by_path:
        :return: None
        """
        result: List[_SkillComponentLoadingItem] = []

        class_index: Dict[
            str, Dict[_SKILL_COMPONENT_TYPES, Set[Type[SkillComponent]]]
        ] = {}
        used_classes = set()
        not_resolved_configurations: Dict[
            Tuple[_SKILL_COMPONENT_TYPES, str], SkillComponentConfiguration
        ] = {}

        # populate indexes
        for _path, component_classes in component_classes_by_path.items():
            for (component_classname, _component_class) in component_classes:
                type_ = self._get_skill_component_type(_component_class)
                class_index.setdefault(component_classname, {}).setdefault(
                    type_, set()
                ).add(_component_class)

        for component_type, by_id in declared_component_classes.items():
            for component_id, component_config in by_id.items():
                path = component_config.file_path
                class_name = component_config.class_name
                if path is not None:
                    classes_in_path = component_classes_by_path[path]
                    component_class_or_none: Optional[Type[SkillComponent]] = next(
                        (
                            actual_class
                            for actual_class_name, actual_class in classes_in_path
                            if actual_class_name == class_name
                        ),
                        None,
                    )
                    enforce(
                        component_class_or_none is not None,
                        self._get_error_message_prefix()
                        + f"Cannot find class '{class_name}' for component '{component_id}' of type '{self._type_to_str(component_type)}' of skill '{self.configuration.public_id}' in module {path}",
                    )
                    component_class = cast(
                        Type[SkillComponent], component_class_or_none
                    )
                    actual_component_type = self._get_skill_component_type(
                        component_class
                    )
                    enforce(
                        actual_component_type == component_type,
                        self._get_error_message_prefix()
                        + f"Found class '{class_name}' for component '{component_id}' of type '{self._type_to_str(component_type)}' of skill '{self.configuration.public_id}' in module {path}, but the expected type was {self._type_to_str(component_type)}, found {self._type_to_str(actual_component_type)} ",
                    )
                    used_classes.add(component_class)
                    result.append(
                        _SkillComponentLoadingItem(
                            component_id,
                            component_config,
                            component_class,
                            component_type,
                        )
                    )
                else:
                    # process the configuration at the end of the loop
                    not_resolved_configurations[
                        (component_type, component_id)
                    ] = component_config

        for (component_type, component_id), component_config in copy(
            not_resolved_configurations
        ).items():
            class_name = component_config.class_name
            classes_by_type = class_index.get(class_name, {})
            enforce(
                class_name in class_index and component_type in classes_by_type,
                self._get_error_message_prefix()
                + f"Cannot find class '{class_name}' for skill component '{component_id}' of type '{self._type_to_str(component_type)}'",
            )
            classes = classes_by_type[component_type]
            not_used_classes = classes.difference(used_classes)
            enforce(
                not_used_classes != 0,
                f"Cannot find class of skill '{self.configuration.public_id}' for component configuration '{component_id}' of type '{self._type_to_str(component_type)}'.",
            )
            enforce(
                len(not_used_classes) == 1,
                self._get_error_message_ambiguous_classes(
                    class_name, not_used_classes, component_type, component_id
                ),
            )
            not_used_class = list(not_used_classes)[0]
            result.append(
                _SkillComponentLoadingItem(
                    component_id, component_config, not_used_class, component_type
                )
            )
            used_classes.add(not_used_class)

        return result

    @classmethod
    def _type_to_str(cls, component_type: _SKILL_COMPONENT_TYPES) -> str:
        """Get the string of a component type."""
        return component_type.__name__.lower()

    def _get_error_message_prefix(self) -> str:
        """Get error message prefix."""
        return f"Error while loading skill '{self.configuration.public_id}': "

    def _get_error_message_ambiguous_classes(
        self,
        class_name: str,
        not_used_classes: Set,
        component_type: _SKILL_COMPONENT_TYPES,
        component_id: str,
    ) -> str:
        return f"{self._get_error_message_prefix()}found many classes with name '{class_name}' for component '{component_id}' of type '{self._type_to_str(component_type)}' in the following modules: {', '.join([c.__module__ for c in not_used_classes])}"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the helpers/serializers module."""
from google.protobuf.struct_pb2 import Struct

from aea.helpers.serializers import DictProtobufStructSerializer


def test_encode_decode_i():
    """Test encode decode logic."""
    case = {
        "key1": True,
        "key2": 0.12,
        "key3": 100,
        "key4": "some string",
        "key5": b"some bytes string",
        "key6": Struct(),
        "_need_patch": {},
    }
    encoded = DictProtobufStructSerializer.encode(case)
    case.pop("_need_patch")
    assert isinstance(encoded, bytes)
    decoded = DictProtobufStructSerializer.decode(encoded)
    assert case == decoded


def test_encode_decode_ii():
    """Test encode decode logic."""
    case = {
        "key1": True,
        "key2": 0.12,
        "key3": 100,
        "key4": "some string",
        "key5": b"some bytes string",
        "key6": {"key1": True, "key2": 0.12},
    }
    encoded = DictProtobufStructSerializer.encode(case)
    assert isinstance(encoded, bytes)
    decoded = DictProtobufStructSerializer.decode(encoded)
    assert case == decoded



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the simple_data_request skill."""
from pathlib import Path

from aea.test_tools.test_skill import BaseSkillTestCase

from tests.conftest import ROOT_DIR


class SimpleDataRequestTestCase(BaseSkillTestCase):
    """Sets the simple_data_request class up for testing (overrides the necessary config values so tests can be done on simple_data_request skill)."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "simple_data_request"
    )

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.mocked_method = "some_method"
        cls.mocked_url = "some_url"
        cls.mocked_shared_state_key = "some_name_for_data"

        config_overrides = {
            "behaviours": {
                "http_request": {
                    "args": {"method": cls.mocked_method, "url": cls.mocked_url}
                }
            },
            "handlers": {
                "http": {"args": {"shared_state_key": cls.mocked_shared_state_key}}
            },
        }

        super().setup(config_overrides=config_overrides)
